/*
(c) 2010 Perttu Ahola <celeron55@gmail.com>
*/

#include "map.h"
//#include "player.h"
#include "main.h"

//TODO: make a struct settings or something
//s16 g_viewing_range_nodes = 50; // fast
s16 g_viewing_range_nodes = 75; // fairly good
//s16 g_viewing_range_nodes = 85;
//s16 g_viewing_range_nodes = 100;


void limitBox(core::aabbox3d<s16> & box, core::aabbox3d<s16> & limits)
{
	if(box.MinEdge.X < limits.MinEdge.X)
		box.MinEdge.X = limits.MinEdge.X;
	if(box.MaxEdge.X > limits.MaxEdge.X)
		box.MaxEdge.X = limits.MaxEdge.X;
	if(box.MinEdge.Y < limits.MinEdge.Y)
		box.MinEdge.Y = limits.MinEdge.Y;
	if(box.MaxEdge.Y > limits.MaxEdge.Y)
		box.MaxEdge.Y = limits.MaxEdge.Y;
	if(box.MinEdge.Z < limits.MinEdge.Z)
		box.MinEdge.Z = limits.MinEdge.Z;
	if(box.MaxEdge.Z > limits.MaxEdge.Z)
		box.MaxEdge.Z = limits.MaxEdge.Z;
}

MapNode * MapBlock::getNodePtrMap(v3s16 p)
{
	if(p.X < 0 || p.X >= size.X
			|| p.Y < 0 || p.Y >= size.Y
			|| p.Z < 0 || p.Z >= size.Z)
	{
		return map->getNodePtrNoCreate(posRelative + p);
	}
	else
	{
		return &data[p.Z*size.X*size.Y + p.Y*size.X + p.X];
	}
}

void MapBlock::setNodeMap(v3s16 p, MapNode & n)
{
	if(p.X < 0 || p.X >= size.X
			|| p.Y < 0 || p.Y >= size.Y
			|| p.Z < 0 || p.Z >= size.Z)
	{
		map->setNode(posRelative + p, n);
	}
	else
	{
		data[p.Z*size.X*size.Y + p.Y*size.X + p.X] = n;
	}
}

/*
	startpos:
	translate_dir: unit vector with only one of x, y or z
	face_dir: unit vector with only one of x, y or z
*/
void MapBlock::updateFastFaceRow(v3s16 startpos,
		u16 length,
		v3s16 translate_dir,
		v3s16 face_dir,
		core::list<FastFace*> &dest)
{
	// The maximum difference in light to tolerate in the same face
	// TODO: make larger?
	//light_t max_light_diff = 0.15;
	light_t max_light_diff = 0.2;

	/*
		Precalculate some variables
	*/
	v3f translate_dir_f(translate_dir.X, translate_dir.Y,
			translate_dir.Z); // floating point conversion
	v3f face_dir_f(face_dir.X, face_dir.Y,
			face_dir.Z); // floating point conversion
	v3f posRelative_f(posRelative.X, posRelative.Y,
			posRelative.Z); // floating point conversion

	v3s16 p = startpos;
	/*
		The light in the air lights the surface is taken from
		the node that is air.
	*/
	light_t light = getFaceLight(p, face_dir);
	
	u16 continuous_materials_count = 0;
	
	u8 material0 = getNodeMaterial(p);
	u8 material1 = getNodeMaterial(p + face_dir);
		
	for(u16 j=0; j<length; j++)
	{
		bool next_is_different = true;
		
		v3s16 p_next;
		u8 material0_next = 0;
		u8 material1_next = 0;
		light_t light_next = 0;
		if(j != length - 1){
			p_next = p + translate_dir;
			material0_next = getNodeMaterial(p_next);
			material1_next = getNodeMaterial(p_next + face_dir);
			light_next = getFaceLight(p_next, face_dir);
			
			if(material0_next == material0
					&& material1_next == material1
					&& fabs(light_next - light) / (light+0.01) < max_light_diff)
			{
				next_is_different = false;
			}
		}

		continuous_materials_count++;
		
		if(next_is_different)
		{
			/*
				If there is a face (= materials differ)
			*/
			if(material0 != material1
					&& material0 != MATERIAL_IGNORE
					&& material1 != MATERIAL_IGNORE){
				// Floating point conversion of the position vector
				v3f pf(p.X, p.Y, p.Z);
				// Center point of face (kind of)
				v3f sp = pf - ((f32)continuous_materials_count / 2. - 0.5) * translate_dir_f;
				v3f scale(1,1,1);
				if(translate_dir.X != 0){
					scale.X = continuous_materials_count;
				}
				if(translate_dir.Y != 0){
					scale.Y = continuous_materials_count;
				}
				if(translate_dir.Z != 0){
					scale.Z = continuous_materials_count;
				}
				FastFace *f;
				// If node at sp is not air
				if(material0 != MATERIAL_AIR){
					f = makeFastFace(material0, light,
							sp, face_dir_f, scale,
							posRelative_f);
				}
				// If node at sp is air
				else{
					f = makeFastFace(material1, light,
							sp+face_dir_f, -1*face_dir_f, scale,
							posRelative_f);
				}
				dest.push_back(f);
			}

			continuous_materials_count = 0;
			material0 = material0_next;
			material1 = material1_next;
			light = light_next;
		}
		
		p = p_next;
	}
}

void MapBlock::updateFastFaces()
{
	v3s16 p = getPosRelative();
	/*std::cout<<"MapBlock("<<p.X<<","<<p.Y<<","<<p.Z<<")"
			<<"::updateFastFaces(): ";*/
			//<<"::updateFastFaces()"<<std::endl;

	core::list<FastFace*> *fastfaces_new = new core::list<FastFace*>;
	
	/*
		These are started and stopped one node overborder to
		take changes in those blocks in account.
		
		FIXME: At the moment they are not, it drops the FPS from 12 to 9.
		FIXME: It can cause a glitch in some rare cases of something changing
		       just beyond the viewing range and then the viewing range
			   extending to there, or something like that.
	*/

	/*
		Go through every y,z and get top faces in rows of x
	*/
	for(s16 y=0; y<size.Y; y++){
	//for(s16 y=-1; y<size.Y; y++){
		for(s16 z=0; z<size.Z; z++){
			updateFastFaceRow(v3s16(0,y,z), size.X,
					v3s16(1,0,0),
					v3s16(0,1,0),
					*fastfaces_new);
		}
	}
	/*
		Go through every x,y and get right faces in rows of z
	*/
	for(s16 x=0; x<size.X; x++){
	//for(s16 x=-1; x<size.X; x++){
		for(s16 y=0; y<size.Y; y++){
			updateFastFaceRow(v3s16(x,y,0), size.Z,
					v3s16(0,0,1),
					v3s16(1,0,0),
					*fastfaces_new);
		}
	}
	/*
		Go through every y,z and get back faces in rows of x
	*/
	for(s16 z=0; z<size.Z; z++){
	//for(s16 z=-1; z<size.Z; z++){
		for(s16 y=0; y<size.Y; y++){
			updateFastFaceRow(v3s16(0,y,z), size.X,
					v3s16(1,0,0),
					v3s16(0,0,1),
					*fastfaces_new);
		}
	}

	/*
		Replace the list
	*/

	core::list<FastFace*> *fastfaces_old = fastfaces;

	fastfaces_mutex.Lock();
	fastfaces = fastfaces_new;
	fastfaces_mutex.Unlock();

	/*std::cout<<" Number of faces: old="<<fastfaces_old->getSize()
			<<" new="<<fastfaces_new->getSize()<<std::endl;*/

	/*
		Clear and dump the old list
	*/

	core::list<FastFace*>::Iterator i = fastfaces_old->begin();
	for(; i != fastfaces_old->end(); i++)
	{
		delete *i;
	}
	fastfaces_old->clear();
	delete fastfaces_old;

	//std::cout<<"added "<<fastfaces.getSize()<<" faces."<<std::endl;
}

MapBlock * Map::getBlockNoCreate(v3s16 p)
{
	static v3s16 p_last(0,0,0);
	static MapBlock *block_last = NULL;

	MapBlock *block;

	if(block_last != NULL && p == p_last){
		return block_last;
	}
	
	// If block doesn't exist, return NULL
	core::map<v3s16, MapBlock*>::Node *n = blocks.find(p);
	if(n == NULL)
	{
		/*
			TODO: Check if block is stored on disk
			      and load dynamically
		*/
		block = NULL;
	}
	// If block exists, return it
	else{
		block = n->getValue();
	}
	
	// Cache the last result
	p_last = p;
	block_last = block;
	
	return block;
}

MapBlock * Map::getBlock(v3s16 p)
{
	/*
		Find out whether block exists; if it doesn't, create one.
		Then return the block.
	*/

	MapBlock *block = getBlockNoCreate(p);

	if(block)
	{
		return block;
	}

	block_create_mutex.Lock();

	/*
		Block doesn't exist
	*/

	/*
	std::cout<<"[Generating MapBlock("
			<<p.X<<","<<p.Y<<","<<p.Z<<")]";
	*/

	/*
		Create a block from scratch
	*/

	block = new MapBlock(this,
			v3s16(MAP_BLOCKSIZE,
				MAP_BLOCKSIZE,
				MAP_BLOCKSIZE),
			v3s16(p.X * MAP_BLOCKSIZE,
				p.Y * MAP_BLOCKSIZE,
				p.Z * MAP_BLOCKSIZE));
	
	blocks.insert(p, block);

	/*
		Generate some content

		TODO: Generate something more interesting
	*/
	if(p.Y == 0){
		block->drawbox(0,0,0, MAP_BLOCKSIZE,1,MAP_BLOCKSIZE, 1);
		//TODO: draw a cone or something
		s16 y=1;
		for(s16 r=MAP_BLOCKSIZE/3 - (((p.X + p.Z) % 2 == 0) ? 5 : 0); r>=1; r--){
			block->drawbox(MAP_BLOCKSIZE/2-r,y,MAP_BLOCKSIZE/2-r,
					r*2,1,r*2, 0);
			y++;
		}
		if((p.X + p.Z) % 2 != 0){
			block->drawbox(MAP_BLOCKSIZE/2-1, 1, 0,
					2, 2, MAP_BLOCKSIZE, MATERIAL_AIR);
			block->drawbox(0, 0, 0,
					4, 1, 4, MATERIAL_AIR);
		}
		/*block->drawbox(3,1,3, 3,1,3, 0);
		block->drawbox(4,2,4, 1,2,1, 0);*/
	}
	else if(p.Y == -1){
		block->drawbox(0,0,0, MAP_BLOCKSIZE,MAP_BLOCKSIZE,
				MAP_BLOCKSIZE, 0);
		block->drawbox(0, MAP_BLOCKSIZE-4, 0,
				4, 3, 4, MATERIAL_AIR);
		block->drawbox(1, MAP_BLOCKSIZE-1, 0,
				2, 1, 4, MATERIAL_AIR);
	}
	else if(p.Y < 0){
		block->drawbox(0,0,0, MAP_BLOCKSIZE,MAP_BLOCKSIZE,
				MAP_BLOCKSIZE, 0);
	}

	block_create_mutex.Unlock();

	return block;
}

/*
	TODO: Check the order of changing lighting and recursing in
	these functions (choose the faster one)

	TODO: It has to done like this: cache the neighbours that were
	changed; then recurse to them.
*/

/*
	Goes recursively through the neighbours of the node.

	Alters only transparent nodes.

	If the lighting of the neighbour is lower than the lighting of
	the node was (before changing it to 0 at the step before), the
	lighting of the neighbour is set to 0 and then the same stuff
	repeats for the neighbour.

	Some things are made strangely to make it as fast as possible.

	Usage:
	core::list<v3s16> light_sources;
	core::map<v3s16, MapBlock*> modified_blocks;
	f32 oldlight = node_at_pos.light;
	node_at_pos.light = 0;
	unLightNeighbors(pos, oldlight, light_sources, modified_blocks);
*/
void Map::unLightNeighbors(v3s16 pos, f32 oldlight,
		core::list<v3s16> & light_sources,
		core::map<v3s16, MapBlock*> & modified_blocks)
{
	v3s16 dirs[6] = {
		v3s16(0,0,1), // back
		v3s16(0,1,0), // top
		v3s16(1,0,0), // right
		v3s16(0,0,-1), // front
		v3s16(0,-1,0), // bottom
		v3s16(-1,0,0), // left
	};
	
	// Loop through 6 neighbors
	for(u16 i=0; i<6; i++){
		// Get the position of the neighbor node
		v3s16 n2pos = pos + dirs[i];
		
		// Get the block where the node is located
		v3s16 blockpos = getNodeBlockPos(n2pos);
		MapBlock *block = getBlockNoCreate(blockpos);

		// If block is nonexistent, retreat.
		if(block == NULL)
			continue;
		
		// Find if block is in cache container
		core::map<v3s16, MapBlock*>::Node *cacheblocknode;
		cacheblocknode = modified_blocks.find(blockpos);
		
		// If the block is not found in the cache
		if(cacheblocknode == NULL)
		{
			/*
				Add block to cache container. It could be a 'set' but
				there is no such thing in Irrlicht.
				
				We can use the pointer as the value of a map just fine,
				it gets nicely cached that way, too.
			*/
			modified_blocks.insert(blockpos, block);
		}
		
		// Calculate relative position in block
		v3s16 relpos = n2pos - blockpos * MAP_BLOCKSIZE;
		// Get node straight from the block (fast!)
		MapNode *n2 = block->getNodePtr(relpos);
		
		/*
			If the neighbor is dimmer than what was specified
			as oldlight (the light of the previous node)...
		*/
		if(n2->light < oldlight - 0.001){
			if(n2->transparent()){
				light_t current_light = n2->light;
				n2->light = 0.0;
				unLightNeighbors(n2pos, current_light,
						light_sources, modified_blocks);
			}
		}
		else{
			light_sources.push_back(n2pos);
		}
	}
}

/*
	Goes recursively through the neighbours of the node.

	Alters only transparent nodes.

	If the lighting of the neighbour is lower than the calculated
	lighting coming from the current node to it, the lighting of
	the neighbour is set and the same thing repeats.

	Some things are made strangely to make it as fast as possible.

	Usage:
	core::map<v3s16, MapBlock*> modified_blocks;
	lightNeighbors(pos, node_at_pos.light, modified_blocks);
*/
void Map::lightNeighbors(v3s16 pos, f32 oldlight,
		core::map<v3s16, MapBlock*> & modified_blocks)
{
	v3s16 dirs[6] = {
		v3s16(0,0,1), // back
		v3s16(0,1,0), // top
		v3s16(1,0,0), // right
		v3s16(0,0,-1), // front
		v3s16(0,-1,0), // bottom
		v3s16(-1,0,0), // left
	};
	
	f32 newlight = oldlight * 0.75;

	core::list<v3s16> neighbour_cache;

	// Loop through 6 neighbors
	for(u16 i=0; i<6; i++){
		// Get the position of the neighbor node
		v3s16 n2pos = pos + dirs[i];
		
		// Get the block where the node is located
		v3s16 blockpos = getNodeBlockPos(n2pos);
		MapBlock *block = getBlockNoCreate(blockpos);

		// If block is nonexistent, retreat.
		if(block == NULL)
			continue;
		
		// Find if block is in cache container
		core::map<v3s16, MapBlock*>::Node *cacheblocknode;
		cacheblocknode = modified_blocks.find(blockpos);
		
		// If the block is not found in the cache
		if(cacheblocknode == NULL)
		{
			/*
				Add block to cache container. It could be a 'set' but
				there is no such thing in Irrlicht.
				
				We can use the pointer as the value of a map just fine,
				it gets nicely cached that way, too.
			*/
			modified_blocks.insert(blockpos, block);
		}
		
		// Calculate relative position in block
		v3s16 relpos = n2pos - blockpos * MAP_BLOCKSIZE;
		// Get node straight from the block (fast!)
		MapNode *n2 = block->getNodePtr(relpos);
		
		/*
			If the neighbor is dimmer than what was specified
			as oldlight (the light of the previous node)...
		*/
		if(n2->light < newlight - 0.001){
			if(n2->transparent()){
				n2->light = newlight;
				// Cache and recurse at last step for maximum speed
				neighbour_cache.push_back(n2pos);
				//lightNeighbors(n2pos, newlight, modified_blocks);
			}
		}
		/*else{
			light_ends.push_back(n2pos);
		}*/
	}

	core::list<v3s16>::Iterator j = neighbour_cache.begin();
	for(; j != neighbour_cache.end(); j++){
		lightNeighbors(*j, newlight, modified_blocks);
	}
}

v3s16 Map::getBrightestNeighbour(v3s16 p)
{
	v3s16 dirs[6] = {
		v3s16(0,0,1), // back
		v3s16(0,1,0), // top
		v3s16(1,0,0), // right
		v3s16(0,0,-1), // front
		v3s16(0,-1,0), // bottom
		v3s16(-1,0,0), // left
	};
	
	f32 brightest_light = -1.0;
	v3s16 brightest_pos(0,0,0);

	// Loop through 6 neighbors
	for(u16 i=0; i<6; i++){
		// Get the position of the neighbor node
		v3s16 n2pos = p + dirs[i];
		MapNode *n2 = getNodePtrNoCreate(n2pos);
		if(n2 == NULL)
			continue;
		if(n2->light > brightest_light){
			brightest_light = n2->light;
			brightest_pos = n2pos;
		}
	}
	if(brightest_light < -0.001)
		throw;
	return brightest_pos;
}

/*
	This is called after changing a node from transparent to opaque.
	The lighting value of the node should be left as-is after changing
	other values. This sets the lighting value to 0.
*/
void Map::nodeAddedUpdate(v3s16 p, f32 lightwas)
{
	std::cout<<"Map::nodeAddedUpdate()"<<std::endl;

	core::list<v3s16> light_sources;
	core::map<v3s16, MapBlock*> modified_blocks;
	//MapNode n = getNode(p);

	/*
		From this node to nodes underneath:
		If lighting is sunlight (1.0), unlight neighbours and
		set lighting to 0.
		Else discontinue.
	*/

	bool node_under_sunlight = true;
	
	v3s16 toppos = p + v3s16(0,1,0);
	MapNode *topnode = getNodePtrNoCreate(toppos);
	
	/*
		If there is a node at top and it doesn't have sunlight,
		there has not been any sunlight going down.

		Otherwise there probably is.
	*/
	if(topnode != NULL)
	{
		if(topnode->light < 0.999)
			node_under_sunlight = false;
	}

	// Add the block of the added node to modified_blocks
	v3s16 blockpos = getNodeBlockPos(p);
	MapBlock *block = getBlockNoCreate(blockpos);
	modified_blocks.insert(blockpos, block);
	
	// Unlight neighbours of node.
	// This means setting light of all consequent dimmer nodes
	// to 0.
	MapNode *n = getNodePtrNoCreate(p);
	if(n == NULL) throw; // Should exist
	unLightNeighbors(p, lightwas, light_sources, modified_blocks);
	n->light = 0;
	
	if(node_under_sunlight)
	{
		s16 y = p.Y - 1;
		for(;; y--){
			std::cout<<"y="<<y<<std::endl;
			v3s16 n2pos(p.X, y, p.Z);
			
			MapNode *n2 = getNodePtrNoCreate(n2pos);
			
			if(n2 == NULL)
				break;
			
			if(n2->light >= 0.999){
				std::cout<<"doing"<<std::endl;
				unLightNeighbors(n2pos, n2->light, light_sources, modified_blocks);
				n2->light = 0;
			}
			else
				break;
		}
	}
	
	core::list<v3s16>::Iterator j = light_sources.begin();
	for(; j != light_sources.end(); j++)
	{
		lightNeighbors(*j, getNode(*j).light, modified_blocks);
	}

	core::map<v3s16, MapBlock*>::Iterator i = modified_blocks.getIterator();
	for(; i.atEnd() == false; i++)
	{
		i.getNode()->getValue()->updateFastFaces();
	}
}

/*
*/
void Map::nodeRemovedUpdate(v3s16 p)
{
	std::cout<<"Map::nodeRemovedUpdate()"<<std::endl;
	
	core::map<v3s16, MapBlock*> modified_blocks;

	bool node_under_sunlight = true;
	
	v3s16 toppos = p + v3s16(0,1,0);
	MapNode *topnode = getNodePtrNoCreate(toppos);
	
	/*
		If there is a node at top and it doesn't have sunlight,
		there will be no sunlight going down.
	*/
	if(topnode != NULL)
	{
		if(topnode->light < 0.999)
			node_under_sunlight = false;
	}

	// Add the block of the removed node to modified_blocks
	v3s16 blockpos = getNodeBlockPos(p);
	MapBlock *block = getBlockNoCreate(blockpos);
	modified_blocks.insert(blockpos, block);
	
	if(node_under_sunlight)
	{
		s16 ybottom = propagateSunlight(p, modified_blocks);
		/*std::cout<<"Node was under sunlight. "
				"Propagating sunlight";
		std::cout<<" -> ybottom="<<ybottom<<std::endl;*/
		s16 y = p.Y;
		for(; y >= ybottom; y--)
		{
			v3s16 p2(p.X, y, p.Z);
			/*std::cout<<"lighting neighbors of node ("
					<<p2.X<<","<<p2.Y<<","<<p2.Z<<")"
					<<std::endl;*/
			lightNeighbors(p2, getNode(p2).light, modified_blocks);
		}
	}
	else
	{
		// Set the lighting of this node to 0
		MapNode *n = getNodePtrNoCreate(p);
		if(n == NULL) throw; // Should exist
		n->light = 0;
	}

	// Get the brightest neighbour node and propagate light from it
	v3s16 n2p = getBrightestNeighbour(p);
	MapNode *n2 = getNodePtrNoCreate(n2p);
	if(n2 == NULL) throw; // Should exist
	lightNeighbors(n2p, n2->light, modified_blocks);
	
	core::map<v3s16, MapBlock*>::Iterator i = modified_blocks.getIterator();
	for(; i.atEnd() == false; i++)
	{
		i.getNode()->getValue()->updateFastFaces();
	}
}

/*
	Propagates sunlight down from a node.
	Starting point gets sunlight.

	Returns the lowest y value of where the sunlight went.
*/
s16 Map::propagateSunlight(v3s16 start,
		core::map<v3s16, MapBlock*> & modified_blocks)
{
	s16 y = start.Y;
	for(; ; y--)
	{
		v3s16 pos(start.X, y, start.Z);
		
		v3s16 blockpos = getNodeBlockPos(pos);
		MapBlock *block = getBlockNoCreate(blockpos);
		if(block == NULL)
			break;

		v3s16 relpos = pos - blockpos*MAP_BLOCKSIZE;
		MapNode *n = block->getNodePtr(relpos);
		//MapNode *n = getNodePtrNoCreate(pos);
		if(n == NULL)
			break;

		if(n->transparent())
		{
			n->light = 1.0;

			modified_blocks.insert(blockpos, block);
		}
		else{
			break;
		}
	}
	return y + 1;
}
		
void Map::updateLighting(core::list<MapBlock*> & a_blocks,
		core::map<v3s16, MapBlock*> & modified_blocks)
{
	std::cout<<"Map::updateLighting(): "
			<<a_blocks.getSize()<<" blocks"<<std::endl;

	std::cout<<"Flooding direct sunlight"<<std::endl;
	
	// Copy list
	core::list<MapBlock*> temp_blocks = a_blocks;

	/*
		Go from the highest blocks to the lowest, propagating sunlight
		through them
	*/

	while(temp_blocks.empty() == false){
		// Get block with highest position in Y
		
		core::list<MapBlock*>::Iterator highest_i = temp_blocks.end();
		v3s16 highest_pos(0,-32768,0);

		core::list<MapBlock*>::Iterator i;
		for(i = temp_blocks.begin(); i != temp_blocks.end(); i++)
		{
			MapBlock *block = *i;
			v3s16 pos = block->getPosRelative();
			if(highest_i == temp_blocks.end() || pos.Y > highest_pos.Y){
				highest_i = i;
				highest_pos = pos;
			}
		}

		// Update sunlight in block
		MapBlock *block = *highest_i;
		block->propagateSunlight();

		// Add to modified blocks
		v3s16 blockpos = getNodeBlockPos(highest_pos);
		modified_blocks.insert(blockpos, block);

		// Remove block from list
		temp_blocks.erase(highest_i);
	}

	// Add other light sources here

	//

	std::cout<<"Spreading light to remaining air"<<std::endl;

	/*
		Spread light to air that hasn't got light
		
		TODO: This shouldn't have to be done to every block
		      Though it doesn't hurt much i guess.
	*/

	core::list<MapBlock*>::Iterator i = a_blocks.begin();
	for(; i != a_blocks.end(); i++)
	{
		MapBlock *block = *i;
		v3s16 pos = block->getPosRelative();
		s16 xmin = pos.X;
		s16 zmin = pos.Z;
		s16 ymin = pos.Y;
		s16 xmax = pos.X + MAP_BLOCKSIZE - 1;
		s16 zmax = pos.Z + MAP_BLOCKSIZE - 1;
		s16 ymax = pos.Y + MAP_BLOCKSIZE - 1;
		for(s16 z=zmin; z<=zmax; z++){
			for(s16 x=xmin; x<=xmax; x++){
				for(s16 y=ymax; y>=ymin; y--){
					v3s16 pos(x, y, z);
					MapNode *n = getNodePtrNoCreate(pos);
					if(n == NULL)
						break;
					if(n->light > NO_LIGHT_MAX){
						v3s16 pos(x,y,z);
						lightNeighbors(pos, n->light, modified_blocks);
					}
				}
			}
		}
		std::cout<<"X";
		std::cout.flush();
	}
	std::cout<<std::endl;
}

core::aabbox3d<s16> Map::getDisplayedBlockArea()
{
	camera_mutex.Lock();
	core::aabbox3d<s16> box_nodes(floatToInt(camera_position));
	camera_mutex.Unlock();

	s16 d = g_viewing_range_nodes;
	box_nodes.MinEdge -= v3s16(d,d,d);
	box_nodes.MaxEdge += v3s16(d,d,d);

	return core::aabbox3d<s16>(
			getNodeBlockPos(box_nodes.MinEdge),
			getNodeBlockPos(box_nodes.MaxEdge));
}

void Map::render(video::IVideoDriver* driver,
	video::SMaterial *materials)
{
	//std::cout<<"Rendering map..."<<std::endl;
	
	core::list<MapBlock*>::Iterator bi;

	/*
		Collect all blocks that are in the view range

		Should not optimize more here as we want to auto-update
		all changed nodes in viewing range at the next step.
	*/
	
	core::list<MapBlock*> blocks_displayed;
	core::aabbox3d<s16> box_blocks = getDisplayedBlockArea();

	for(s16 y=box_blocks.MaxEdge.Y; y>=box_blocks.MinEdge.Y; y--){
		for(s16 z=box_blocks.MinEdge.Z; z<=box_blocks.MaxEdge.Z; z++){
			for(s16 x=box_blocks.MinEdge.X; x<=box_blocks.MaxEdge.X; x++)
			{
				/*
					Add block to list
				*/
				MapBlock *block = getBlockNoCreate(v3s16(x,y,z));
				if(block == NULL)
					continue;
				blocks_displayed.push_back(block);
			}
		}
	}

	u8 material_in_use;
	material_in_use = MATERIAL_AIR;

	if(blocks_displayed.getSize() == 0)
		return;
	
	/*
		Draw all displayed faces
	*/

	u32 facecount = 0;
	
	for(bi = blocks_displayed.begin(); bi != blocks_displayed.end(); bi++){
		MapBlock *block = *bi;
		
		/*
			Compare block position to camera position, skip
			if not seen on display
		*/
#if 1
		v3s16 blockpos_nodes = block->getPosRelative();

		// Block center position
		v3f blockpos(
				((float)blockpos_nodes.X + MAP_BLOCKSIZE/2) * BS,
				((float)blockpos_nodes.Y + MAP_BLOCKSIZE/2) * BS,
				((float)blockpos_nodes.Z + MAP_BLOCKSIZE/2) * BS
		);

		camera_mutex.Lock();

		// Block position relative to camera
		v3f blockpos_relative = blockpos - camera_position;
		//v3f blockpos_relative = blockpos - v3f(0,0,0);

		// Distance in camera direction (+=front, -=back)
		f32 dforward = blockpos_relative.dotProduct(camera_direction);
		//f32 dforward = blockpos_relative.dotProduct(v3f(0,0,1));

		camera_mutex.Unlock();

		// Total distance
		f32 d = blockpos_relative.getLength();

		// Cosine of the angle between the camera direction
		// and the block direction (camera_direction is an unit vector)
		f32 cosangle = dforward / blockpos_relative.getLength();
		// Compensate for the size of the block
		// (as the block has to be shown even if it's a bit off FOV)
		// This is an estimate.
		cosangle += (1.44*1.44*MAP_BLOCKSIZE*BS/2.0) / dforward;
		//cosangle += (2.0*MAP_BLOCKSIZE) / d;

		/*// If block is on the backside of camera, skip it
		if(d < -1.0 * MAP_BLOCKSIZE * BS)
			continue;*/
			
		// If block is not in the field of view, skip it
		//if(cosangle < cos(FOV_ANGLE/2) && d > 1.44*MAP_BLOCKSIZE*BS/2.0)
		if(cosangle < cos(FOV_ANGLE/2) && d > 1.44*1.44*MAP_BLOCKSIZE*BS/2.0)
			continue;
#endif
		block->fastfaces_mutex.Lock();

		core::list<FastFace*>::Iterator i =
				block->fastfaces->begin();

		for(; i != block->fastfaces->end(); i++)
		{
			FastFace *f = *i;
			
			if(f->material != material_in_use){
				driver->setMaterial(materials[f->material]);
				material_in_use = f->material;
			}


			u16 indices[] = {0,1,2,3};
			
			driver->drawVertexPrimitiveList(f->vertices, 4, indices, 1,
					video::EVT_STANDARD, scene::EPT_QUADS, video::EIT_16BIT);

			facecount++;
		}

		block->fastfaces_mutex.Unlock();
	}
	
	//std::cout<<"Rendered "<<facecount<<" faces"<<std::endl;
	
	//std::cout<<"done"<<std::endl;
}

/*
	Returns true if updated something
*/
bool Map::updateChangedVisibleArea()
{
	core::aabbox3d<s16> box_blocks = getDisplayedBlockArea();
	core::list<MapBlock*> blocks_displayed;

	for(s16 y=box_blocks.MaxEdge.Y; y>=box_blocks.MinEdge.Y; y--){
		for(s16 z=box_blocks.MinEdge.Z; z<=box_blocks.MaxEdge.Z; z++){
			for(s16 x=box_blocks.MinEdge.X; x<=box_blocks.MaxEdge.X; x++)
			{
				// This intentionally creates new blocks on demand
				MapBlock *block = getBlock(v3s16(x,y,z));
				blocks_displayed.push_back(block);
			}
		}
	}

	if(blocks_displayed.getSize() == 0){
		return false;
	}
	
	/*
		TODO: Update closest ones first

		Update the lighting and the faces of changed blocks that are
		to be displayed.
	*/
	
	core::list<MapBlock*> blocks_changed;

	core::list<MapBlock*>::Iterator bi;
	for(bi = blocks_displayed.begin(); bi != blocks_displayed.end(); bi++)
	{
		MapBlock *block = *bi;
		if(block->getChangedFlag())
		{
			if(blocks_changed.empty() == true){
				// Print initial message when first is found
				std::cout<<"Updating changed MapBlocks at ";
			}

			v3s16 p = block->getPosRelative();
			std::cout<<"("<<p.X<<","<<p.Y<<","<<p.Z<<") ";
			std::cout.flush();

			blocks_changed.push_back(block);
			block->resetChangedFlag();
		}
	}

	// Quit if nothing has changed
	if(blocks_changed.empty()){
		return false;
	}

	std::cout<<std::endl;
	
	core::map<v3s16, MapBlock*> modified_blocks;

	std::cout<<"Updating lighting"<<std::endl;

	updateLighting(blocks_changed, modified_blocks);

	std::cout<<"Updating face cache of changed blocks"<<std::endl;

	core::map<v3s16, MapBlock*>::Iterator i = modified_blocks.getIterator();
	for(; i.atEnd() == false; i++)
	{
		i.getNode()->getValue()->updateFastFaces();
		std::cout<<"X";
		std::cout.flush();
	}

	std::cout<<std::endl;

	return true;
}

void * MapUpdateThread::Thread()
{
	if(map == NULL)
		return NULL;
		
	ThreadStarted();

	while(getRun())
	{
		//std::cout<<"UpdateThread running"<<std::endl;

		map->updateChangedVisibleArea();
		
		sleep(1);
	}

	std::cout<<"UpdateThread stopped"<<std::endl;

	return NULL;
}

