/*
(c) 2010 Perttu Ahola <celeron55@gmail.com>
*/

#include "heightmap.h"

/*
	source: Connects borders with this
	blocksize: Size of heightmap
	offset_blocks: placement of block on master
	master: source for corner heights
*/
void Heightmap::GenerateContinued(HeightmapSource *source, s16 blocksize,
		v2s16 offset_blocks, Heightmap *master, f32 default_value,
		f32 randmax, f32 randfactor)
{
	// Works only with blocksize=2,4,8,16,32,64,...
	s16 a = blocksize;
	
	// Check that a is a power of two
	if((a & (a-1)) != 0)
		throw;
		
	allocate(a+1, a+1);

	/*
		Seed corner values from master heightmap or default value
	*/

	v2s16 corners[4] = {
		v2s16(0,0),
		v2s16(1,0),
		v2s16(0,1),
		v2s16(1,1)
	};
	for(u16 i=0; i<4; i++){
		f32 h;
		if(master == NULL){
			h = default_value;
		}
		else{
			h = master->getGroundHeight(offset_blocks + corners[i]);
			if(h < GROUNDHEIGHT_VALID_MINVALUE)
				h = default_value;
		}
		seedNode(corners[i] * a, h);
	}

	/*std::cout<<"Seeded heightmap:"<<std::endl;
	print();*/

	setSource(source, blocksize, offset_blocks);

	DiamondSquare(randmax, randfactor);
}

