#include "mapsector.h"

MapBlock * MapSector::getBlockBuffered(s16 y)
{
	static s16 y_last = 0;
	static MapBlock *block_last = NULL;

	MapBlock *block;

	if(block_last != NULL && y == y_last){
		return block_last;
	}
	
	// If block doesn't exist, return NULL
	core::map<s16, MapBlock*>::Node *n = m_blocks.find(y);
	if(n == NULL)
	{
		/*
			TODO: Check if block is stored on disk
			      and load dynamically
		*/
		block = NULL;
	}
	// If block exists, return it
	else{
		block = n->getValue();
	}
	
	// Cache the last result
	y_last = y;
	block_last = block;
	
	return block;
}

Ref<MapBlock> MapSector::getBlock(s16 y)
{
	/*std::cout<<"MapSector("<<m_pos.X<<","<<m_pos.Y<<")::getBlock("<<y
			<<")"<<std::endl;*/
	
	//m_blocks_mutex.Lock();

	MapBlock *block = getBlockBuffered(y);

	if(block == NULL)
		//return Ref<MapBlock>(NULL);
		return NULL; // This is actually same as the latter

	Ref<MapBlock> ref(block);

	//m_blocks_mutex.Unlock();

	/*std::cout<<"MapSector("<<m_pos.X<<","<<m_pos.Y<<")::GetBlock(): "
			<<"ref.get()->getRefcount()="
			<<ref.get()->getRefcount()
			<<std::endl;*/

	return ref;
}

Ref<MapBlock> MapSector::createBlankBlock(s16 y)
{
	/*std::cout<<"MapSector("<<m_pos.X<<","<<m_pos.Y<<")::createBlankBlock("<<y
			<<")"<<std::endl;*/

	//m_blocks_mutex.Lock();

	assert(getBlockBuffered(y) == NULL);

	v3s16 blockpos_map(m_pos.X, y, m_pos.Y);
	
	MapBlock *block = new MapBlock(m_parent, blockpos_map);
	
	m_blocks.insert(y, block);

	Ref<MapBlock> ref(block);

	//m_blocks_mutex.Unlock();

	return ref;
}

core::list< Ref<MapBlock> > MapSector::getBlocks()
{
	core::list< Ref<MapBlock> > ref_list;

	core::map<s16, MapBlock*>::Iterator bi;

	bi = m_blocks.getIterator();
	for(; bi.atEnd() == false; bi++)
	{
		MapBlock *b = bi.getNode()->getValue();
		ref_list.push_back(Ref<MapBlock>(b));
	}
	
	return ref_list;
}

f32 MapSector::getGroundHeight(v2s16 p)
{
	//TODO
	return 0.0;
}

void MapSector::setGroundHeight(v2s16 p, f32 h)
{
	//TODO
	assert(0);
}

//END
