/*
(c) 2010 Perttu Ahola <celeron55@gmail.com>

Minetest

TODO:

- Storing map on disk, preferably dynamically

- struct settings, with a mutex and get/set functions

- Torches or similar light sources

- A menu

- Map autogeneration

TODO: Make unit tests for heightmap stuff
TODO: Heightmaps to refer to parent in all relevant cases

TODO: A cache class that can be used with lightNeighbors,
      unlightNeighbors and probably many others. Look for
	  implementation in lightNeighbors

TODO: Check for obsolete todos

TODO: Add mutexes to necessary places

*/

#include <iostream>
#include <time.h>

#include "common_irrlicht.h"

/*
	Setting this to 1 enables a special camera mode that forces
	the renderers to think that the camera statically points from
	the starting place to a static direction.

	This allows one to move around with the player and see what
	is actually drawn behind solid things etc.
*/
#define FIELD_OF_VIEW_TEST 0

#define ENABLE_TESTS 1

#ifdef _MSC_VER
#pragma comment(lib, "Irrlicht.lib")
#pragma comment(lib, "jthread.lib")
// This would get rid of the console window
//#pragma comment(linker, "/subsystem:windows /ENTRY:mainCRTStartup")
#endif

#include "map.h"
#include "player.h"
#include "main.h"
#include "test.h"

const char *g_material_filenames[MATERIALS_COUNT] =
{
	"../data/stone.png",
	"../data/grass.png",
};

s16 g_viewing_range_nodes = 50; // fast
//s16 g_viewing_range_nodes = 60; // fairly good
//s16 g_viewing_range_nodes = 75; // fairly good
//s16 g_viewing_range_nodes = 100;

class CSampleSceneNode : public scene::ISceneNode
{
	core::aabbox3d<f32> box;
	video::SMaterial *materials;
	Map *map;

public:

	CSampleSceneNode(scene::ISceneNode* parent, scene::ISceneManager* mgr, s32 id,
			Map *a_map, video::SMaterial *a_materials)
		: scene::ISceneNode(parent, mgr, id)
	{
		map = a_map;
		materials = a_materials;
		
		/*box = core::aabbox3d<f32>(0,0,0,
				map->getW()*BS, map->getH()*BS, map->getD()*BS);*/
		/*box = core::aabbox3d<f32>(0,0,0,
				map->getSizeNodes().X * BS,
				map->getSizeNodes().Y * BS,
				map->getSizeNodes().Z * BS);*/
		box = core::aabbox3d<f32>(-BS*1000000,-BS*1000000,-BS*1000000,
				BS*1000000,BS*1000000,BS*1000000);
	}

	virtual void OnRegisterSceneNode()
	{
		if (IsVisible)
			SceneManager->registerNodeForRendering(this);

		ISceneNode::OnRegisterSceneNode();
	}

	virtual void render()
	{
		video::IVideoDriver* driver = SceneManager->getVideoDriver();
		driver->setTransform(video::ETS_WORLD, AbsoluteTransformation);
		map->render(driver, materials);
	}
	
	virtual const core::aabbox3d<f32>& getBoundingBox() const
	{
		//TODO: Get a value from the map
		return box;
	}

	/*virtual u32 getMaterialCount() const
	{
		return 1;
	}

	virtual video::SMaterial& getMaterial(u32 i)
	{
		return materials[0];
	}*/	
};

class MyEventReceiver : public IEventReceiver
{
public:
	// This is the one method that we have to implement
	virtual bool OnEvent(const SEvent& event)
	{
		// Remember whether each key is down or up
		if(event.EventType == irr::EET_KEY_INPUT_EVENT)
				keyIsDown[event.KeyInput.Key] = event.KeyInput.PressedDown;

		if(event.EventType == irr::EET_MOUSE_INPUT_EVENT)
		{
			if(event.MouseInput.Event == EMIE_LMOUSE_PRESSED_DOWN)
			{
				leftclicked = true;
			}
			if(event.MouseInput.Event == EMIE_RMOUSE_PRESSED_DOWN)
			{
				rightclicked = true;
			}
		}

		return false;
	}

	// This is used to check whether a key is being held down
	virtual bool IsKeyDown(EKEY_CODE keyCode) const
	{
		return keyIsDown[keyCode];
	}

	MyEventReceiver()
	{
		for (u32 i=0; i<KEY_KEY_CODES_COUNT; ++i)
				keyIsDown[i] = false;
		leftclicked = false;
		rightclicked = false;
	}

	bool leftclicked;
	bool rightclicked;
private:
	// We use this array to store the current state of each key
	bool keyIsDown[KEY_KEY_CODES_COUNT];
	//s32 mouseX;
	//s32 mouseY;
};

/*
static void test()
{
	core::map<v3s16, s16> m;
	m.insert(v3s16(0,0,0), 0);
	m.insert(v3s16(0,0,1), 1);
	m.insert(v3s16(0,2,2), 2);
	m.insert(v3s16(0,0,3), 3);

	std::cout<<(int)m.find(v3s16(0,0,3))<<std::endl;
	std::cout<<(int)m.find(v3s16(0,1,0))<<std::endl;
	std::cout<<m[v3s16(0,0,3)]<<std::endl; //3
	std::cout<<m[v3s16(0,2,2)]<<std::endl; //2
	std::cout<<m[v3s16(1,0,0)]<<std::endl; //-
}
*/

int main()
{
	/*
		Test stuff
	*/

	//test();
	//return 0;
	/*TestThread thread;
	thread.Start();
	std::cout<<"thread started"<<std::endl;
	while(thread.IsRunning()) sleep(1);
	std::cout<<"thread ended"<<std::endl;
	return 0;*/
	
	if(ENABLE_TESTS)
		run_tests();
		
	/*
		Initialization
	*/

	srand(time(0));

	MyEventReceiver receiver;

	video::E_DRIVER_TYPE driverType;

	/*printf("Please select the driver you want for this example:\n"\
		" (a) OpenGL 1.5\n (b) Direct3D 9.0c\n (c) Direct3D 8.1\n"\
		" (d) Burning's Software Renderer\n (e) Software Renderer\n"\
		" (f) NullDevice\n (otherKey) exit\n\n");

	char i;
	std::cin >> i;

	switch(i)
	{
		case 'a': driverType = video::EDT_OPENGL;   break;
		case 'b': driverType = video::EDT_DIRECT3D9;break;
		case 'c': driverType = video::EDT_DIRECT3D8;break;
		case 'd': driverType = video::EDT_BURNINGSVIDEO;break;
		case 'e': driverType = video::EDT_SOFTWARE; break;
		case 'f': driverType = video::EDT_NULL;     break;
		default: return 1;
	}*/

	// create device and exit if creation failed

	u16 screenW = 640;
	u16 screenH = 480;

#ifdef _WIN32
	//driverType = video::EDT_DIRECT3D9; // Doesn't seem to work
	driverType = video::EDT_OPENGL;
#else
	driverType = video::EDT_OPENGL;
#endif

	IrrlichtDevice *device;
	device = createDevice(driverType,
			core::dimension2d<u32>(screenW, screenH),
			16, false, false, false, &receiver);

	if (device == 0)
		return 1; // could not create selected driver.

	video::IVideoDriver* driver = device->getVideoDriver();
	// These make the textures not to show at all
	//driver->setTextureCreationFlag(video::ETCF_ALWAYS_16_BIT);
	//driver->setTextureCreationFlag(video::ETCF_OPTIMIZED_FOR_SPEED );

	scene::ISceneManager* smgr = device->getSceneManager();

	gui::IGUIEnvironment* guienv = device->getGUIEnvironment();
	gui::IGUISkin* skin = guienv->getSkin();
	gui::IGUIFont* font = guienv->getFont("../data/fontlucida.png");
	if(font)
		skin->setFont(font);
	//skin->setColor(gui::EGDC_BUTTON_TEXT, video::SColor(255,0,0,0));
	skin->setColor(gui::EGDC_BUTTON_TEXT, video::SColor(255,255,255,255));
	//skin->setColor(gui::EGDC_3D_HIGH_LIGHT, video::SColor(0,0,0,0));
	//skin->setColor(gui::EGDC_3D_SHADOW, video::SColor(0,0,0,0));
	skin->setColor(gui::EGDC_3D_HIGH_LIGHT, video::SColor(255,0,0,0));
	skin->setColor(gui::EGDC_3D_SHADOW, video::SColor(255,0,0,0));
	
	const wchar_t *text = L"Loading...";
	core::vector2d<s32> center(screenW/2, screenH/2);
	core::dimension2d<u32> textd = font->getDimension(text);
	std::cout<<"Text w="<<textd.Width<<" h="<<textd.Height<<std::endl;
	// Have to add a bit to disable the text from word wrapping
	//core::vector2d<s32> textsize(textd.Width+4, textd.Height);
	core::vector2d<s32> textsize(300, textd.Height);
	core::rect<s32> textrect(center - textsize/2, center + textsize/2);

	gui::IGUIStaticText *gui_loadingtext = guienv->addStaticText(
			text, textrect, false, false);
	gui_loadingtext->setTextAlignment(gui::EGUIA_CENTER, gui::EGUIA_UPPERLEFT);

	driver->beginScene(true, true, video::SColor(255,0,0,0));
	guienv->drawAll();
	driver->endScene();

	video::SMaterial materials[MATERIALS_COUNT];
	for(u16 i=0; i<MATERIALS_COUNT; i++){
		//materials[i].Wireframe = false;
		//materials[i].MaterialType = video::EMT_SOLID; //default
		//materials[i].MaterialType = video::EMT_ONETEXTURE_BLEND;
		//materials[i].Lighting = true;
		materials[i].Lighting = false;
		//materials[i].AmbientColor = video::SColor(255,255,255,255); //default
		//materials[i].DiffuseColor = video::SColor(255,255,255,255); //default
		//materials[i].EmissiveColor = video::SColor(255,255,255,255);
		//materials[i].EmissiveColor = video::SColor(0,0,0,0); //default
		//materials[i].ColorMaterial = video::ECM_DIFFUSE; //default
		//materials[i].AntiAliasing = false;
		materials[i].BackfaceCulling = false;
		//materials[i].FrontfaceCulling = false;
		//materials[i].GouraudShading = false;
		//materials[i].setTexture(0, driver->getTexture("../data/test16.png"));
		//materials[i].setTexture(0, driver->getTexture("../data/test16_noborder.png"));
		const char *filename = g_material_filenames[i];
		if(filename != NULL){
			video::ITexture *t = driver->getTexture(filename);
			if(t == NULL){
				std::cout<<"Texture could not be loaded: \""
						<<filename<<"\""<<std::endl;
				return 1;
			}
			materials[i].setTexture(0, driver->getTexture(filename));
		}
		//materials[i].setFlag(video::EMF_TEXTURE_WRAP, video::ETC_REPEAT);
		materials[i].setFlag(video::EMF_BILINEAR_FILTER, false);
		//materials[i].setFlag(video::EMF_ANISOTROPIC_FILTER, false);
	}

	std::cout<<"Creating player and map"<<std::endl;

	/*
		The map.
	*/
	Map map;

	map.StartUpdater();

#if 1
	while(map.status.getReady() == false){
		wchar_t *text = map.status.getNiceText();
		gui_loadingtext->setText(text);
		free(text);

		driver->beginScene(true, true, video::SColor(255,0,0,0));
		guienv->drawAll();
		driver->endScene();
	}
#endif

	// Now when the map is ready, we are ready to put a player in

	Player player;
	//player.position = v3f(MAP_BLOCKSIZE/2*BS-BS/2, MAP_BLOCKSIZE/2*BS, MAP_BLOCKSIZE/2*BS-BS/2);
	//player.position = v3f(0, 51*BS, 0);
	player.position = v3f(0, BS*2 + map.getGroundHeight(v2s16(0,0)), 0);
	player.speed = v3f(0,0,0);

	/*
		Create the scene node that draws the map
	*/

	/*CSampleSceneNode *node =*/
	new CSampleSceneNode(smgr->getRootSceneNode(), smgr, 666, &map, materials);
	/*if (node)
		node->setPosition(v3f(-BS*40, -BS*2, -BS*40));*/
	
	/*
		Create the camera node
	*/

	scene::ICameraSceneNode* camera = smgr->addCameraSceneNode(
		0, // Camera parent
		v3f(BS*100, BS*2, BS*100), // Look from
		v3f(BS*100+1, BS*2, BS*100), // Look to
		-1 // Camera ID
   	);

	if(camera == NULL)
		return 1;
	
	camera->setFOV(FOV_ANGLE);
	// Just so big a value that everything rendered is visible
	camera->setFarValue(BS*1000);
	
	f32 camera_yaw = 0; // "right/left"
	f32 camera_pitch = 0; // "up/down"
	
	// Random constants
	f32 walkspeed_max = WALKSPEED_MAX;
	f32 walk_acceleration = WALK_ACCELERATION;
	
	/*
		The mouse cursor needs not be visible, so we hide it via the
		irr::IrrlichtDevice::ICursorControl.
	*/
	device->getCursorControl()->setVisible(false);
	
	gui_loadingtext->remove();
	
	//gui::IGUIStaticText *positiontextgui = 
	guienv->addStaticText(
			L"Minetest-c55", core::rect<s32>(5, 5, 5+100, 5+textsize.Y),
			false, false);
	

	/*
		Main loop
	*/

	bool first_loop_after_window_activation = true;
	
	s32 lastFPS = -1;
	
	// Time is in milliseconds
	u32 lasttime = device->getTimer()->getTime();
	
	while(device->run())
	{
		// Collected during the loop and displayed
		core::list< core::aabbox3d<f32> > hilightboxes;
		
		/*
			Special keys
		*/
		if(receiver.IsKeyDown(irr::KEY_ESCAPE))
		{
			break;
		}

		/*
			Player speed control
		*/

		v3f move_direction = v3f(0,0,1);
		move_direction.rotateXZBy(camera_yaw);
		
		v3f speed = v3f(0,0,0);
		if(receiver.IsKeyDown(irr::KEY_KEY_W))
		{
			speed += move_direction;
		}
		if(receiver.IsKeyDown(irr::KEY_KEY_S))
		{
			speed -= move_direction;
		}
		if(receiver.IsKeyDown(irr::KEY_KEY_A))
		{
			speed += move_direction.crossProduct(v3f(0,1,0));
		}
		if(receiver.IsKeyDown(irr::KEY_KEY_D))
		{
			speed += move_direction.crossProduct(v3f(0,-1,0));
		}
		if(receiver.IsKeyDown(irr::KEY_SPACE))
		{
			if(player.touching_ground){
				//player_speed.Y = 30*BS;
				//player.speed.Y = 5*BS;
				player.speed.Y = 6.5*BS;
			}
		}

		// The speed of the player (Y is ignored)
		speed = speed.normalize() * walkspeed_max;

		
		/*
			Time difference calculation
		*/
		u32 time = device->getTimer()->getTime();
		f32 dtime_total; // in seconds
		if(time > lasttime)
			dtime_total = (time - lasttime) / 1000.0;
		else
			dtime_total = 0;
		lasttime = time;
		
		// Limit the stuff happening in a single frame
		if(dtime_total > 0.2)
			dtime_total = 0.2;

		// Maximum time increment (for collision detection etc)
		// Allow 0.1 blocks per increment
		// time = distance / speed
		f32 dtime_max_increment = 0.1*BS / player.speed.getLength();
		if(dtime_max_increment > 0.01)
			dtime_max_increment = 0.01;

		/*
			Stuff that has a maximum time increment:
		*/
		do
		{
			f32 dtime;
			if(dtime_total > dtime_max_increment)
				dtime = dtime_max_increment;
			else
				dtime = dtime_total;
			dtime_total -= dtime;

			//std::cout<<"dtime="<<dtime<<std::endl;

			/*
				Collision detection
			*/
			
			f32 inc = walk_acceleration * BS * dtime;
			
			if(player.speed.X < speed.X - inc)
				player.speed.X += inc;
			else if(player.speed.X > speed.X + inc)
				player.speed.X -= inc;
			else if(player.speed.X < speed.X)
				player.speed.X = speed.X;
			else if(player.speed.X > speed.X)
				player.speed.X = speed.X;

			if(player.speed.Z < speed.Z - inc)
				player.speed.Z += inc;
			else if(player.speed.Z > speed.Z + inc)
				player.speed.Z -= inc;
			else if(player.speed.Z < speed.Z)
				player.speed.Z = speed.Z;
			else if(player.speed.Z > speed.Z)
				player.speed.Z = speed.Z;

			//player.speed.Y -= 9.81*BS*dtime;
			player.speed.Y -= 9.81*BS*dtime*2;
			//player.speed.Y -= 2*BS*dtime;

			player.move(dtime, map);
	
			/*std::cout<<"player.touching_ground="
					<<player.touching_ground<<std::endl;*/
		}
		while(dtime_total > 0.001);

		/*
			Mouse and camera control
		*/
		
		if(device->isWindowActive())
		{
			if(first_loop_after_window_activation){
				first_loop_after_window_activation = false;
			}
			else{
				s32 dx = device->getCursorControl()->getPosition().X - 320;
				s32 dy = device->getCursorControl()->getPosition().Y - 240;
				camera_yaw -= dx*0.2;
				camera_pitch += dy*0.2;
				if(camera_pitch < -89.9) camera_pitch = -89.9;
				if(camera_pitch > 89.9) camera_pitch = 89.9;
			}
			device->getCursorControl()->setPosition(320, 240);
		}
		else{
			first_loop_after_window_activation = true;
		}
		
		v3f camera_direction = v3f(0,0,1);
		camera_direction.rotateYZBy(camera_pitch);
		camera_direction.rotateXZBy(camera_yaw);

		v3f camera_position =
				player.position + v3f(0, BS+BS/2, 0);

		camera->setPosition(camera_position);
		camera->setTarget(camera_position + camera_direction);

		if(FIELD_OF_VIEW_TEST){
			map.updateCamera(v3f(0,0,0), v3f(0,0,1));
		}
		else{
			map.updateCamera(camera_position, camera_direction);
		}
		
		/*
			Calculate what block is the crosshair pointing to
		*/
		
		//u32 t1 = device->getTimer()->getTime();
		
		f32 d = 4; // max. distance
		core::line3d<f32> shootline(camera_position,
				camera_position + camera_direction * BS * (d+1));
		
		bool nodefound = false;
		v3s16 nodepos;
		v3s16 neighbourpos;
		core::aabbox3d<f32> nodefacebox;
		f32 mindistance = BS * 1001;
		
		v3s16 pos_i = Map::floatToInt(player.position);

		s16 a = d;
		s16 ystart = pos_i.Y + 0 - (camera_direction.Y<0 ? a : 1);
		s16 zstart = pos_i.Z - (camera_direction.Z<0 ? a : 1);
		s16 xstart = pos_i.X - (camera_direction.X<0 ? a : 1);
		s16 yend = pos_i.Y + 1 + (camera_direction.Y>0 ? a : 1);
		s16 zend = pos_i.Z + (camera_direction.Z>0 ? a : 1);
		s16 xend = pos_i.X + (camera_direction.X>0 ? a : 1);
		
		for(s16 y = ystart; y <= yend; y++){
			for(s16 z = zstart; z <= zend; z++){
				for(s16 x = xstart; x <= xend; x++){
					try{
						if(map.getNode(x,y,z).d == MATERIAL_AIR){
							continue;
						}
					}catch(InvalidPositionException &e){
						continue;
					}

					v3s16 np(x,y,z);
					v3f npf = Map::intToFloat(np);
					
					f32 d = 0.01;
					
					v3s16 directions[6] = {
						v3s16(0,0,1), // back
						v3s16(0,1,0), // top
						v3s16(1,0,0), // right
						v3s16(0,0,-1),
						v3s16(0,-1,0),
						v3s16(-1,0,0),
					};

					for(u16 i=0; i<6; i++){
					//{u16 i=3;
						v3f dir_f = v3f(directions[i].X,
								directions[i].Y, directions[i].Z);
						v3f centerpoint = npf + dir_f * BS/2;
						f32 distance =
								(centerpoint - camera_position).getLength();
						
						if(distance < mindistance){
							//std::cout<<"for centerpoint=("<<centerpoint.X<<","<<centerpoint.Y<<","<<centerpoint.Z<<"): distance < mindistance"<<std::endl;
							//std::cout<<"npf=("<<npf.X<<","<<npf.Y<<","<<npf.Z<<")"<<std::endl;
							core::CMatrix4<f32> m;
							m.buildRotateFromTo(v3f(0,0,1), dir_f);

							// This is the back face
							v3f corners[2] = {
								v3f(BS/2, BS/2, BS/2),
								v3f(-BS/2, -BS/2, BS/2+d)
							};
							
							for(u16 j=0; j<2; j++){
								m.rotateVect(corners[j]);
								corners[j] += npf;
								//std::cout<<"box corners["<<j<<"]: ("<<corners[j].X<<","<<corners[j].Y<<","<<corners[j].Z<<")"<<std::endl;
							}

							//core::aabbox3d<f32> facebox(corners[0],corners[1]);
							core::aabbox3d<f32> facebox(corners[0]);
							facebox.addInternalPoint(corners[1]);

							if(facebox.intersectsWithLine(shootline)){
								nodefound = true;
								nodepos = np;
								neighbourpos = np + directions[i];
								mindistance = distance;
								nodefacebox = facebox;
							}
						}
					}
				}
			}
		}

		if(nodefound){
			//std::cout<<"nodefound == true"<<std::endl;
			//std::cout<<"nodepos=("<<nodepos.X<<","<<nodepos.Y<<","<<nodepos.Z<<")"<<std::endl;
			//std::cout<<"neighbourpos=("<<neighbourpos.X<<","<<neighbourpos.Y<<","<<neighbourpos.Z<<")"<<std::endl;

			static v3s16 nodepos_old(-1,-1,-1);
			if(nodepos != nodepos_old){
				std::cout<<"Pointing at ("<<nodepos.X<<","
						<<nodepos.Y<<","<<nodepos.Z<<")"<<std::endl;
				nodepos_old = nodepos;

				/*wchar_t positiontext[20];
				swprintf(positiontext, 20, L"(%i,%i,%i)",
						nodepos.X, nodepos.Y, nodepos.Z);
				positiontextgui->setText(positiontext);*/
			}

			hilightboxes.push_back(nodefacebox);
			
			if(receiver.leftclicked){
				std::cout<<"Removing block (MapNode)"<<std::endl;
				MapNode n;
				n.d = MATERIAL_AIR;
				map.setNode(nodepos, n);
				map.nodeRemovedUpdate(nodepos);
			}
			if(receiver.rightclicked){
				std::cout<<"Placing block (MapNode)"<<std::endl;
				f32 light = map.getNode(neighbourpos).light;
				MapNode n;
				n.d = 0;
				map.setNode(neighbourpos, n);
				map.nodeAddedUpdate(neighbourpos, light);
			}
		}
		else{
			//std::cout<<"nodefound == false"<<std::endl;
			//positiontextgui->setText(L"");
		}

		receiver.leftclicked = false;
		receiver.rightclicked = false;

		/*
			Drawing begins
		*/

		driver->beginScene(true, true, video::SColor(255,90,140,200));

		//std::cout<<"smgr->drawAll()"<<std::endl;

		smgr->drawAll();

		core::vector2d<s32> displaycenter(screenW/2,screenH/2);
		driver->draw2DLine(displaycenter - core::vector2d<s32>(10,0),
				displaycenter + core::vector2d<s32>(10,0),
				video::SColor(255,255,255,255));
		driver->draw2DLine(displaycenter - core::vector2d<s32>(0,10),
				displaycenter + core::vector2d<s32>(0,10),
				video::SColor(255,255,255,255));
		
		video::SMaterial m;
		m.Thickness = 10;
		m.Lighting = false;
		driver->setMaterial(m);

		for(core::list< core::aabbox3d<f32> >::Iterator i=hilightboxes.begin();
				i != hilightboxes.end(); i++){
			driver->draw3DBox(*i, video::SColor(255,0,0,0));
		}

		guienv->drawAll();

		driver->endScene();

		/*
			Drawing ends
		*/

		u16 fps = driver->getFPS();

		if (lastFPS != fps)
		{
			core::stringw str = L"Minetest [";
			str += driver->getName();
			str += "] FPS:";
			str += fps;

			device->setWindowCaption(str.c_str());
			lastFPS = fps;
		}


		/*}
		else
			device->yield();*/
	}

	/*
	In the end, delete the Irrlicht device.
	*/
	device->drop();
	return 0;
}

//END
