/*
(c) 2010 Perttu Ahola <celeron55@gmail.com>
*/

#ifndef MAPSECTOR_HEADER
#define MAPSECTOR_HEADER

#include <jmutex.h>
#include "common_irrlicht.h"
#include "mapblock.h"
#include "heightmap.h"
#include "exceptions.h"

/*
	This is an Y-wise stack of MapBlocks.
	
	TODO:
	- Add heightmap functionality
	- Implement node container functionality and modify Map to use it
*/

class AlreadyExistsException : public std::exception
{
	virtual const char * what() const throw()
	{
		return "AlreadyExistsException";
	}
};

class MapSector :
		public NodeContainer,
		public Heightmappish
{
private:
	
	// The pile of MapBlocks
	core::map<s16, MapBlock*> m_blocks;
	//JMutex m_blocks_mutex; // For public access functions

	NodeContainer *m_parent;
	// Position on parent (in MapBlock widths)
	v2s16 m_pos;

	MapBlock *getBlockBuffered(s16 y);

	FixedHeightmap *m_heightmap;

	// Be sure to set this to NULL when the cached block is deleted 
	MapBlock *m_block_cache;
	s16 m_block_cache_y;

public:

	//JMutex mutex;

	MapSector(NodeContainer *parent, v2s16 pos, Heightmap * masterheightmap):
			m_parent(parent),
			m_pos(pos),
			m_heightmap(NULL),
			m_block_cache(NULL)
	{
		//mutex.Init();
		m_heightmap = new FixedHeightmap(masterheightmap, pos, MAP_BLOCKSIZE);
	}

	~MapSector()
	{
		//TODO: clear m_blocks
		core::map<s16, MapBlock*>::Iterator i = m_blocks.getIterator();
		for(; i.atEnd() == false; i++)
		{
			delete i.getNode()->getValue();
		}

		delete m_heightmap;
	}

	FixedHeightmap * getHeightmap()
	{
		if(m_heightmap == NULL)
			throw TargetInexistentException();
		return m_heightmap;
	}

	// If these return !NULL, the returned block should be
	// dropped() after use
	MapBlock * getBlock(s16 y);
	MapBlock * createBlankBlock(s16 y);
	
	core::list<MapBlock*> getBlocks();
	
	f32 getGroundHeight(v2s16 p, bool generate=false);
	void setGroundHeight(v2s16 p, f32 y, bool generate=false);
};

#endif

