#include "mapsector.h"

MapBlock * MapSector::getBlockBuffered(s16 y)
{
	MapBlock *block;

	if(m_block_cache != NULL && y == m_block_cache_y){
		return m_block_cache;
	}
	
	// If block doesn't exist, return NULL
	core::map<s16, MapBlock*>::Node *n = m_blocks.find(y);
	if(n == NULL)
	{
		/*
			TODO: Check if block is stored on disk
			      and load dynamically
		*/
		block = NULL;
	}
	// If block exists, return it
	else{
		block = n->getValue();
	}
	
	// Cache the last result
	m_block_cache_y = y;
	m_block_cache = block;
	
	return block;
}

MapBlock * MapSector::getBlock(s16 y)
{
	/*std::cout<<"MapSector("<<m_pos.X<<","<<m_pos.Y<<")::getBlock("<<y
			<<")"<<std::endl;*/
	
	//m_blocks_mutex.Lock();

	MapBlock *block = getBlockBuffered(y);

	if(block == NULL)
		throw InvalidPositionException();
		//return Ref<MapBlock>(NULL);
		//return NULL; // This is actually same as the latter
	
	return block;
	
	//MapBlock * ref(block);

	//m_blocks_mutex.Unlock();

	/*std::cout<<"MapSector("<<m_pos.X<<","<<m_pos.Y<<")::GetBlock(): "
			<<"ref.get()->getRefcount()="
			<<ref.get()->getRefcount()
			<<std::endl;*/

	//return ref;
}

MapBlock * MapSector::createBlankBlock(s16 y)
{
	/*std::cout<<"MapSector("<<m_pos.X<<","<<m_pos.Y<<")::createBlankBlock("<<y
			<<")"<<std::endl;*/

	//m_blocks_mutex.Lock();
	
	// There should not be a block at this position
	if(getBlockBuffered(y) != NULL)
		throw AlreadyExistsException();

	v3s16 blockpos_map(m_pos.X, y, m_pos.Y);
	
	MapBlock *block = new MapBlock(m_parent, blockpos_map);
	
	m_blocks.insert(y, block);

	return block;

	//MapBlock * ref(block);

	//m_blocks_mutex.Unlock();

	//return ref;
}

core::list<MapBlock*> MapSector::getBlocks()
{
	core::list<MapBlock*> ref_list;

	core::map<s16, MapBlock*>::Iterator bi;

	bi = m_blocks.getIterator();
	for(; bi.atEnd() == false; bi++)
	{
		MapBlock *b = bi.getNode()->getValue();
		ref_list.push_back(b);
	}
	
	return ref_list;
}

f32 MapSector::getGroundHeight(v2s16 p, bool generate)
{
	if(m_heightmap == NULL)
		return GROUNDHEIGHT_NOTFOUND_SETVALUE;
	return m_heightmap->getGroundHeight(p);
}

void MapSector::setGroundHeight(v2s16 p, f32 y, bool generate)
{
	if(m_heightmap == NULL)
		return;
	m_heightmap->setGroundHeight(p, y);
}

//END
