#include "mapsector.h"

MapBlock * MapSector::getBlockBuffered(s16 y)
{
	MapBlock *block;

	if(m_block_cache != NULL && y == m_block_cache_y){
		return m_block_cache;
	}
	
	// If block doesn't exist, return NULL
	core::map<s16, MapBlock*>::Node *n = m_blocks.find(y);
	if(n == NULL)
	{
		/*
			TODO: Check if block is stored on disk
			      and load dynamically
		*/
		block = NULL;
	}
	// If block exists, return it
	else{
		block = n->getValue();
	}
	
	// Cache the last result
	m_block_cache_y = y;
	m_block_cache = block;
	
	return block;
}

MapBlock * MapSector::getBlockNoCreate(s16 y)
{
	/*std::cout<<"MapSector("<<m_pos.X<<","<<m_pos.Y<<")::getBlock("<<y
			<<")"<<std::endl;*/
	
	//m_blocks_mutex.Lock();

	MapBlock *block = getBlockBuffered(y);

	if(block == NULL)
		throw InvalidPositionException();
		//return Ref<MapBlock>(NULL);
		//return NULL; // This is actually same as the latter
	
	return block;
	
	/*std::cout<<"MapSector("<<m_pos.X<<","<<m_pos.Y<<")::GetBlock(): "
			<<"ref.get()->getRefcount()="
			<<ref.get()->getRefcount()
			<<std::endl;*/
}

MapBlock * MapSector::createBlankBlockNoInsert(s16 y)
{
	// There should not be a block at this position
	if(getBlockBuffered(y) != NULL)
		throw AlreadyExistsException();

	v3s16 blockpos_map(m_pos.X, y, m_pos.Y);
	
	MapBlock *block = new MapBlock(m_parent, blockpos_map);
	
	return block;
}

MapBlock * MapSector::createBlankBlock(s16 y)
{
	/*std::cout<<"MapSector("<<m_pos.X<<","<<m_pos.Y<<")::createBlankBlock("<<y
			<<")"<<std::endl;*/

	//m_blocks_mutex.Lock();
	
	// There should not be a block at this position
	if(getBlockBuffered(y) != NULL)
		throw AlreadyExistsException();

	MapBlock *block = createBlankBlockNoInsert(y);
	
	m_blocks.insert(y, block);

	return block;
}

/*
	This will generate a new block as needed.

	A valid heightmap is assumed to exist already.
*/
MapBlock * MapSector::getBlock(s16 block_y)
{
	MapBlock *block = getBlockBuffered(block_y);
	if(block != NULL)
		return block;
		
	block = createBlankBlockNoInsert(block_y);

	/*
		TODO: After generating a heightmap, the heightmaps of the
		      neighboring sectors might have been changed. Update them.
			  NOTE: No they couldn't have been - it hasn't been
			        implemented yet!
	*/

	u8 material = 1;

	for(s16 z0=0; z0<MAP_BLOCKSIZE; z0++){
		for(s16 x0=0; x0<MAP_BLOCKSIZE; x0++){
			s16 target_y = getGroundHeight(v2s16(x0,z0));

			if(getSlope(v2s16(x0,z0)).getLength() > 1.05)
				material = 0;
			else
				material = 1;

			for(s16 y0=0; y0<MAP_BLOCKSIZE; y0++){
				s16 real_y = block_y * MAP_BLOCKSIZE + y0;
				MapNode n;
				if(real_y <= target_y)
					n.d = material;
				else
					n.d = MATERIAL_AIR;
				block->setNode(v3s16(x0,y0,z0), n);
			}
		}
	}

	// Apply sunlight to the block
	block->propagateSunlight();
	
	// Insert to container
	m_blocks.insert(block_y, block);

	return block;
}

core::list<MapBlock*> MapSector::getBlocks()
{
	core::list<MapBlock*> ref_list;

	core::map<s16, MapBlock*>::Iterator bi;

	bi = m_blocks.getIterator();
	for(; bi.atEnd() == false; bi++)
	{
		MapBlock *b = bi.getNode()->getValue();
		ref_list.push_back(b);
	}
	
	return ref_list;
}

f32 MapSector::getGroundHeight(v2s16 p, bool generate)
{
	if(m_heightmap == NULL)
		return GROUNDHEIGHT_NOTFOUND_SETVALUE;
	return m_heightmap->getGroundHeight(p);
}

void MapSector::setGroundHeight(v2s16 p, f32 y, bool generate)
{
	if(m_heightmap == NULL)
		return;
	m_heightmap->setGroundHeight(p, y);
}

//END
