#include "environment.h"

Environment::Environment(Map *map)
{
	m_map = map;
}

Environment::~Environment()
{
	// Deallocate players
	for(core::list<Player*>::Iterator i = m_players.begin();
			i != m_players.end(); i++)
	{
		delete (*i);
	}
}

void Environment::step(float dtime)
{
	f32 maximum_player_speed = 0.001; // just some small value
	for(core::list<Player*>::Iterator i = m_players.begin();
			i != m_players.end(); i++)
	{
		f32 speed = (*i)->speed.getLength();
		if(speed > maximum_player_speed)
			maximum_player_speed = speed;
	}
	
	// Maximum time increment (for collision detection etc)
	// Allow 0.1 blocks per increment
	// time = distance / speed
	f32 dtime_max_increment = 0.1*BS / maximum_player_speed;
	// Maximum time increment is 10ms or lower
	if(dtime_max_increment > 0.01)
		dtime_max_increment = 0.01;

	/*
		Stuff that has a maximum time increment
	*/
	do
	{
		f32 dtime_part;
		if(dtime > dtime_max_increment)
			dtime_part = dtime_max_increment;
		else
			dtime_part = dtime;
		dtime -= dtime_part;
		
		/*
			Move players
		*/
		for(core::list<Player*>::Iterator i = m_players.begin();
				i != m_players.end(); i++)
		{
			Player *player = *i;
			player->speed.Y -= 9.81 * BS * dtime_part * 2;
			player->move(dtime_part, *m_map);
		}
	}
	while(dtime > 0.001);

}

Map & Environment::getMap()
{
	return *m_map;
}

void Environment::addPlayer(Player *player)
{
	m_players.push_back(player);
}

void Environment::removePlayer(Player *player)
{
	//TODO
	throw;
}

