#include "server.h"
#include "utility.h"
#include <iostream>
#include "clientserver.h"
#include "map.h"

Server::Server():
	m_env(new ServerMap()),
	m_con(PROTOCOL_ID)
{
}

Server::~Server()
{
}

void Server::bind(unsigned short port)
{
	m_con.Serve(port);
}

void Server::step(float dtime)
{
	m_env.step(dtime);
	m_con.RunTimeouts(dtime);
	for(;;){
		u32 data_maxsize = 1000;
		u8 data[data_maxsize];
		u16 peer_id;
		u16 seqnumber;
		u32 datasize;
		try{
			datasize = m_con.Receive(peer_id, data, data_maxsize, seqnumber);
			ProcessData(data, datasize, peer_id, seqnumber);
		}
		catch(con::NoIncomingDataException &e)
		{
			break;
		}
	}
}

void Server::ProcessData(u8 *data, u32 datasize, u16 peer_id, u16 seqnumber)
{
	if(datasize < 2)
		return;
	ToServerCommand command = (ToServerCommand)readU16(&data[0]);
	switch(command)
	{
	case TOSERVER_GETBLOCK:{
		//throw NotImplementedException("");

		// Check for too short data
		if(datasize < 8)
			return;
		std::cout<<"Server::ProcessData(): GETBLOCK"
				<<std::endl;
		/*
			Get block data and send it
		*/
		v3s16 p;
		p.X = readS16(&data[2]);
		p.Y = readS16(&data[4]);
		p.Z = readS16(&data[6]);
		MapBlock *block = m_env.getMap().getBlock(p);
	
		u32 replysize = 8 + MapBlock::serializedLength();
		u8 reply[replysize];
		writeU16(&reply[0], TOCLIENT_BLOCKDATA);
		writeS16(&reply[2], p.X);
		writeS16(&reply[4], p.Y);
		writeS16(&reply[6], p.Z);
		block->serialize(&reply[8]);
		// Send as unreliable
		m_con.Send(peer_id, reply, replysize, false);
		}break;
	case TOSERVER_REMOVENODE:{
		if(datasize < 8)
			return;
			
		v3s16 p;
		p.X = readS16(&data[2]);
		p.Y = readS16(&data[4]);
		p.Z = readS16(&data[6]);

		MapNode n;
		n.d = MATERIAL_AIR;
		m_env.getMap().setNode(p, n);

		u32 replysize = 8;
		u8 reply[replysize];
		writeU16(&reply[0], TOCLIENT_REMOVENODE);
		writeS16(&reply[2], p.X);
		writeS16(&reply[4], p.Y);
		writeS16(&reply[6], p.Z);
		// Send as reliable
		// TODO: Send to every client
		m_con.Send(peer_id, reply, replysize, true);
		}break;
	case TOSERVER_ADDNODE:{
		if(datasize < 8 + MapNode::serializedLength())
			return;

		v3s16 p;
		p.X = readS16(&data[2]);
		p.Y = readS16(&data[4]);
		p.Z = readS16(&data[6]);

		MapNode n;
		n.deSerialize(&data[8]);
		m_env.getMap().setNode(p, n);

		u32 replysize = 8 + MapNode::serializedLength();
		u8 reply[replysize];
		writeU16(&reply[0], TOCLIENT_ADDNODE);
		writeS16(&reply[2], p.X);
		writeS16(&reply[4], p.Y);
		writeS16(&reply[6], p.Z);
		n.serialize(&reply[8]);
		// Send as reliable
		// TODO: Send to every client
		m_con.Send(peer_id, reply, replysize, true);
	}break;
	default:
		std::cout<<"Server::ProcessData(): Ingoring unknown command "
				<<command<<std::endl;
	}
}


