/*
(c) 2010 Perttu Ahola <celeron55@gmail.com>

Minetest

SUGGESTION: add a second lighting value to the MS nibble of param of
	air to tell how bright the air node is when there is no sunlight.
	When day changes to night, these two values can be interpolated.
TODO: A menu
TODO: A cache class that can be used with lightNeighbors,
      unlightNeighbors and probably many others. Look for
	  implementation in lightNeighbors
	  - No?
TODO: See if changing to 32-bit position variables doesn't raise
      memory consumption a lot.
TODO: Add more heightmap points to MapSectors
      - Is this needed? The maps are quite good already
TODO: Rendering far map to a texture and drawing it as a background
      - Could be useful it could be done in pieces
TODO: ESC Pause mode in which the cursor is not kept at the center of window.
TODO: Stop player if focus of window is taken away (go to pause mode)
TODO: Optimize and fix makeFastFace or whatever it's called
      - Face calculation is the source of CPU usage on the client
TODO: Set server to automatically find a good spawning place in some
      place where there is water and land.
      - Then try setting the map generator to a more mountainous
	    value
		- I think the map generator needs some way to vary the
		  roughness and mountainousness in different places before
		  this is sensible
SUGGESTION: The client will calculate and send lighting changes and
  the server will randomly check some of them and kick the client out
  if it fails to calculate them right.
  - Actually, it could just start ignoring them and calculate them
    itself.
TODO: Make client fetch stuff mainly from the viewing direction
TODO: Running mode for free mode

TODO: Lighting still sucks!

Doing now:
======================================================================

TODO: Trees
TODO: - Wood block
TODO: - Leaves block

======================================================================

*/

/*
	Setting this to 1 enables a special camera mode that forces
	the renderers to think that the camera statically points from
	the starting place to a static direction.

	This allows one to move around with the player and see what
	is actually drawn behind solid things etc.
*/
#define FIELD_OF_VIEW_TEST 0

// Enable unit tests
#define ENABLE_TESTS 1

#ifdef _MSC_VER
#pragma comment(lib, "Irrlicht.lib")
#pragma comment(lib, "jthread.lib")
// This would get rid of the console window
//#pragma comment(linker, "/subsystem:windows /ENTRY:mainCRTStartup")
#endif

#ifdef _WIN32
	#define WIN32_LEAN_AND_MEAN
	#include <windows.h>
	#define sleep_ms(x) Sleep(x)
#else
	#include <unistd.h>
	#define sleep_ms(x) usleep(x*1000)
#endif

#include <iostream>
#include <fstream>
#include <time.h>
#include <jmutexautolock.h>
#include "common_irrlicht.h"
#include "map.h"
#include "player.h"
#include "main.h"
#include "test.h"
#include "environment.h"
#include "server.h"
#include "client.h"
#include "serialization.h"
#include "constants.h"
#include "strfnd.h"

const char *g_material_filenames[MATERIALS_COUNT] =
{
	"../data/stone.png",
	"../data/grass.png",
	"../data/water.png",
	NULL,
	"../data/tree.png",
	"../data/leaves.png"
};

video::SMaterial g_materials[MATERIALS_COUNT];


JMutex g_range_mutex;
// Blocks are generated in this range from the player
// This is limited vertically to half by Client::fetchBlocks()
s16 g_forcedfetch_range_nodes = FORCEDFETCH_RANGE;
// Blocks are viewed in this range from the player
s16 g_viewing_range_nodes = VIEWING_RANGE_NODES_DEFAULT;
// If true, the preceding value has no meaning and all blocks
// already existing in memory are drawn
bool g_viewing_range_all = false;

/*
	Settings.
	These are loaded from the config file.
*/
float g_wanted_fps = FPS_DEFAULT;
s16 g_viewing_range_nodes_max = VIEWING_RANGE_NODES_MAX;
std::string g_screenW;
std::string g_screenH;
std::string g_dedicated_server;
std::string g_host_game;
std::string g_port;
std::string g_address;
std::string g_name;

/*
	Random stuff
*/

u16 g_selected_material = 0;

bool g_esc_pressed = false;

/*
	Debug streams
	- use these to disable or enable outputs of parts of the program
*/

/*class Nullbuf : public std::streambuf {
	int overflow(int c) { return c; }
};*/
class Nullstream : public std::ostream {
public:
	Nullstream():
		std::ostream(0)
	{
	}
private:
};

#define DEBUGFILE "debug.txt"


std::ofstream dfile(DEBUGFILE);
Nullstream dummyout;

// Connection
//std::ostream *dout_con_ptr = &dfile;
std::ostream *dout_con_ptr = &dummyout;
//std::ostream *derr_con_ptr = &std::cerr;
std::ostream *derr_con_ptr = &dfile;

// Server
std::ostream *dout_server_ptr = &dfile;
std::ostream *derr_server_ptr = &std::cerr;

// Client
std::ostream *dout_client_ptr = &dfile;
std::ostream *derr_client_ptr = &std::cerr;

/*
	Config stuff
*/

bool is_yes(std::string s)
{
	s = trim(s);
	if(s == "y" || s == "yes"
	|| s == "Y" || s == "YES" || s == "Yes"
	|| s == "true" || s == "True" || s == "true")
		return true;
	return false;
}

// Returns false on EOF
bool parseConfigObject(std::istream &is)
{
	// float g_wanted_fps
	// s16 g_viewing_range_nodes_max

	if(is.eof())
		return false;
	
	std::string line;
	std::getline(is, line);
	//std::cout<<"got line: \""<<line<<"\""<<std::endl;

	std::string trimmedline = trim(line);

	//std::cout<<"trimmedline=\""<<trimmedline<<"\""<<std::endl;

	Strfnd sf(trim(line));

	std::string name = sf.next("=");
	name = trim(name);

	if(name == "")
		return true;
	
	std::string value = sf.next("\n");
	value = trim(value);

	std::cout<<"Config name=\""<<name<<"\" value=\""
			<<value<<"\""<<std::endl;
	
	if(name == "wanted_fps")
	{
		g_wanted_fps = atof(value.c_str());
	}
	else if(name == "viewing_range_nodes_max")
	{
		g_viewing_range_nodes_max = atoi(value.c_str());
	}
	else if(name=="screenW")
		g_screenW = value;
	else if(name=="screenH")
		g_screenH = value;
	else if(name == "dedicated_server")
	{
		g_dedicated_server = value;
	}
	else if(name == "host_game")
	{
		g_host_game = value;
	}
	else if(name == "port")
	{
		g_port = value;
	}
	else if(name == "address")
	{
		g_address = value;
	}
	else if(name == "name")
	{
		g_name = value;
	}
	else
	{
		std::cerr<<"Unknown option in config file: \""
				<<line<<"\""<<std::endl;
	}

	return true;
}

void readConfigFile()
{
	const char *filenames[2] =
	{
		"../minetest.conf",
		"../../minetest.conf"
	};

	for(u32 i=0; i<2; i++)
	{
		std::ifstream is(filenames[i]);
		if(is.good() == false)
		{
			std::cout<<DTIME<<"Configuration file not found: "
					<<filenames[i]<<std::endl;
			continue;
		}

		std::cout<<DTIME<<"Parsing configuration file: "
				<<filenames[i]<<std::endl;
				
		while(parseConfigObject(is));
		
		break;
	}
}

/*
	Timestamp stuff
*/

JMutex g_timestamp_mutex;
//std::string g_timestamp;

std::string getTimestamp()
{
	if(g_timestamp_mutex.IsInitialized()==false)
		return "";
	JMutexAutoLock lock(g_timestamp_mutex);
	//return g_timestamp;
	time_t t = time(NULL);
	struct tm *tm = localtime(&t);
	char cs[20];
	strftime(cs, 20, "%H:%M:%S", tm);
	return cs;
}

class MyEventReceiver : public IEventReceiver
{
public:
	// This is the one method that we have to implement
	virtual bool OnEvent(const SEvent& event)
	{
		// Remember whether each key is down or up
		if(event.EventType == irr::EET_KEY_INPUT_EVENT)
		{
			keyIsDown[event.KeyInput.Key] = event.KeyInput.PressedDown;
			
			if(event.KeyInput.PressedDown)
			{
				if(event.KeyInput.Key == irr::KEY_ESCAPE)
				{
					std::cout<<DTIME<<"ESC pressed"<<std::endl;
					g_esc_pressed = true;
				}

				// Material selection
				if(event.KeyInput.Key == irr::KEY_KEY_F)
				{
					if(g_selected_material < USEFUL_MATERIAL_COUNT-1)
						g_selected_material++;
					else
						g_selected_material = 0;
					std::cout<<DTIME<<"Selected material: "
							<<g_selected_material<<std::endl;
				}

				// Viewing range selection
				if(event.KeyInput.Key == irr::KEY_KEY_R)
				{
					JMutexAutoLock lock(g_range_mutex);
					if(g_viewing_range_all)
					{
						g_viewing_range_all = false;
						std::cout<<DTIME<<"Disabled full viewing range"<<std::endl;
					}
					else
					{
						g_viewing_range_all = true;
						std::cout<<DTIME<<"Enabled full viewing range"<<std::endl;
					}
				}
			}
		}

		if(event.EventType == irr::EET_MOUSE_INPUT_EVENT)
		{
			if(event.MouseInput.Event == EMIE_LMOUSE_PRESSED_DOWN)
			{
				leftclicked = true;
			}
			if(event.MouseInput.Event == EMIE_RMOUSE_PRESSED_DOWN)
			{
				rightclicked = true;
			}
		}

		return false;
	}

	// This is used to check whether a key is being held down
	virtual bool IsKeyDown(EKEY_CODE keyCode) const
	{
		return keyIsDown[keyCode];
	}

	MyEventReceiver()
	{
		for (u32 i=0; i<KEY_KEY_CODES_COUNT; ++i)
				keyIsDown[i] = false;
		leftclicked = false;
		rightclicked = false;
	}

	bool leftclicked;
	bool rightclicked;
private:
	// We use this array to store the current state of each key
	bool keyIsDown[KEY_KEY_CODES_COUNT];
	//s32 mouseX;
	//s32 mouseY;
};

void updateViewingRange(f32 frametime, Client *client)
{
#if 1
	static float frametime_avg = 1.0/FPS_DEFAULT;
	frametime_avg = frametime_avg * 0.95 + frametime * 0.05;

	static f32 counter = 0;
	if(counter > 0){
		counter -= frametime;
		return;
	}
	//counter = 5.0; //seconds
	//counter = 1.0; //seconds
	//counter = 1.0; //seconds
	counter = 0.5; //seconds

	float fps_min = g_wanted_fps / 1.3;
	float fps_max = g_wanted_fps * 1.3;

	JMutexAutoLock lock(g_range_mutex);
	
	bool changed = false;
	
	float fraction = sqrt(frametime_avg / (1.0/g_wanted_fps));
	
	if(frametime_avg > 1.0/fps_min
			|| g_viewing_range_nodes > VIEWING_RANGE_NODES_MAX*2)
	{
		if(g_viewing_range_nodes > VIEWING_RANGE_NODES_MIN
				// Don't decrease if using range_max
				&& g_viewing_range_all == false)
		{
			changed = true;
			//g_viewing_range_nodes -= MAP_BLOCKSIZE/4;
			//g_viewing_range_nodes -= g_viewing_range_nodes/10;
			s16 n = (float)g_viewing_range_nodes / fraction;
			if(n < VIEWING_RANGE_NODES_MIN)
				n = VIEWING_RANGE_NODES_MIN;
			g_viewing_range_nodes = n;
		}
	}
	else if(frametime_avg < 1.0/fps_max
			|| g_viewing_range_nodes < VIEWING_RANGE_NODES_MIN)
	{
		if(g_viewing_range_nodes < VIEWING_RANGE_NODES_MAX)
		{
			if(client->isFetchingBlocks() == false)
			{
				changed = true;
				//g_viewing_range_nodes += MAP_BLOCKSIZE/4;
				//g_viewing_range_nodes += g_viewing_range_nodes/10 + 1;
				s16 n = (float)g_viewing_range_nodes / fraction;
				if(n > VIEWING_RANGE_NODES_MAX)
					n = VIEWING_RANGE_NODES_MAX;
				g_viewing_range_nodes = n;
			}
		}
	}

	if(changed)
	{
		std::cout<<DTIME<<"g_viewing_range_nodes = "
				<<g_viewing_range_nodes<<std::endl;
	}
#endif
}

s16 temp16;
f32 tempf;
v3f tempv3f1;
v3f tempv3f2;

void SpeedTests(IrrlichtDevice *device)
{
	/*
		Test stuff
	*/

	//test();
	//return 0;
	/*TestThread thread;
	thread.Start();
	std::cout<<DTIME<<"thread started"<<std::endl;
	while(thread.IsRunning()) sleep(1);
	std::cout<<DTIME<<"thread ended"<<std::endl;
	return 0;*/

	{
		std::cout<<DTIME<<"Testing floating-point conversion speed"<<std::endl;
		u32 time1 = device->getTimer()->getRealTime();
		tempf = 0.001;
		for(u32 i=0; i<10000000; i++){
			temp16 += tempf;
			tempf += 0.001;
		}
		u32 time2 = device->getTimer()->getRealTime();
		u32 fp_conversion_time = time2 - time1;
		std::cout<<DTIME<<"Done. "<<fp_conversion_time<<"ms"<<std::endl;
		//assert(fp_conversion_time < 1000);
	}
	
	{
		std::cout<<DTIME<<"Testing floating-point vector speed"<<std::endl;
		u32 time1 = device->getTimer()->getRealTime();

		tempv3f1 = v3f(1,2,3);
		tempv3f2 = v3f(4,5,6);
		for(u32 i=0; i<40000000; i++){
			tempf += tempv3f1.dotProduct(tempv3f2);
			tempv3f2 += v3f(7,8,9);
		}

		u32 time2 = device->getTimer()->getRealTime();
		u32 dtime = time2 - time1;
		std::cout<<DTIME<<"Done. "<<dtime<<"ms"<<std::endl;
	}

	{
		std::cout<<DTIME<<"Testing core::map speed"<<std::endl;
		u32 time1 = device->getTimer()->getRealTime();
		
		core::map<v2s16, f32> map1;
		tempf = -324;
		for(s16 y=0; y<500; y++){
			for(s16 x=0; x<500; x++){
				map1.insert(v2s16(x,y), tempf);
				tempf += 1;
			}
		}
		for(s16 y=500-1; y>=0; y--){
			for(s16 x=0; x<500; x++){
				tempf = map1[v2s16(x,y)];
			}
		}

		u32 time2 = device->getTimer()->getRealTime();
		u32 dtime = time2 - time1;
		std::cout<<DTIME<<"Done. "<<dtime<<"ms"<<std::endl;
	}

	{
		std::cout<<DTIME<<"Testing mutex speed"<<std::endl;
		u32 time1 = device->getTimer()->getRealTime();
		u32 time2 = time1;
		
		JMutex m;
		m.Init();
		u32 n = 0;
		u32 i = 0;
		do{
			n += 10000;
			for(; i<n; i++){
				m.Lock();
				m.Unlock();
			}
			time2 = device->getTimer()->getRealTime();
		}
		// Do at least 10ms
		while(time2 < time1 + 10);

		u32 dtime = time2 - time1;
		u32 per_ms = n / dtime;
		std::cout<<DTIME<<"Done. "<<dtime<<"ms, "
				<<per_ms<<"/ms"<<std::endl;
	}

	//assert(0);
}

int main()
{
	std::cout<<DTIME<<"minetest-c55"
			" with SER_FMT_VER_HIGHEST="<<(int)SER_FMT_VER_HIGHEST
			<<", ENABLE_TESTS="<<ENABLE_TESTS
			<<std::endl;

	try
	{
	
	// Initialize sockets
	sockets_init();
	atexit(sockets_cleanup);
	
	// Initialize timestamp mutex
	g_timestamp_mutex.Init();

	/*
		Run unit tests
	*/
	if(ENABLE_TESTS)
		run_tests();
	
	//return 0; //DEBUG
		
	/*
		Initialization
	*/

	// Read config file
	readConfigFile();

	// Initialize random seed
	srand(time(0));

	g_range_mutex.Init();
	assert(g_range_mutex.IsInitialized());

	MyEventReceiver receiver;

	/*
		Ask some stuff
	*/

	std::cout<<std::endl<<std::endl;
	char templine[100];
	
	std::cout<<"Dedicated server? [y = yes]: ";
	if(g_dedicated_server != "")
	{
		std::cout<<g_dedicated_server<<std::endl;
		snprintf(templine, 100, "%s", g_dedicated_server.c_str());
	}
	else
	{
		std::cin.getline(templine, 100);
	}

	bool dedicated = false;
	if(templine[0] == 'y')
		dedicated = true;
	if(dedicated)
		std::cout<<"-> yes"<<std::endl;
	else
		std::cout<<"-> no"<<std::endl;
	
	std::cout<<"Port [empty=30000]: ";
	if(g_port != "")
	{
		std::cout<<g_port<<std::endl;
		snprintf(templine, 100, "%s", g_port.c_str());
	}
	else
	{
		std::cin.getline(templine, 100);
	}
	unsigned short port;
	if(templine[0] == 0)
		port = 30000;
	else
		port = atoi(templine);
	std::cout<<"-> "<<port<<std::endl;
	
	if(dedicated)
	{
		std::cout<<std::endl;
		std::cout<<"========================"<<std::endl;
		std::cout<<"Running dedicated server"<<std::endl;
		std::cout<<"========================"<<std::endl;
		std::cout<<std::endl;

		dout_server_ptr = &std::cout;
		
		Server server("../map");
		server.start(port);
	
		for(;;)
		{
			// This is kind of a hack but can be done like this
			// because server.step() is very light
			sleep_ms(100);
			server.step(0.1);

			static int counter = 0;
			counter--;
			if(counter <= 0)
			{
				counter = 10;

				core::list<PlayerInfo> list = server.getPlayerInfo();
				core::list<PlayerInfo>::Iterator i;
				static u32 sum_old = 0;
				u32 sum = PIChecksum(list);
				if(sum != sum_old)
				{
					std::cout<<DTIME<<"Player info:"<<std::endl;
					for(i=list.begin(); i!=list.end(); i++)
					{
						i->PrintLine(&std::cout);
					}
				}
				sum_old = sum;
			}
		}

		return 0;
	}

	bool hosting = false;
	char connect_name[100] = "";

	std::cout<<"Address to connect to [empty = host a game]: ";
	if(g_address != "" && is_yes(g_host_game) == false)
	{
		std::cout<<g_address<<std::endl;
		snprintf(connect_name, 100, "%s", g_address.c_str());
	}
	else
	{
		std::cin.getline(connect_name, 100);
	}
	
	if(connect_name[0] == 0){
		snprintf(connect_name, 100, "127.0.0.1");
		hosting = true;
	}
	
	if(hosting)
		std::cout<<"-> hosting"<<std::endl;
	else
		std::cout<<"-> "<<connect_name<<std::endl;
	
	char playername[PLAYERNAME_SIZE] = "";
	if(g_name != "")
	{
		snprintf(playername, PLAYERNAME_SIZE, "%s", g_name.c_str());
	}
	else
	{
		std::cout<<"Name of player: ";
		std::cin.getline(playername, PLAYERNAME_SIZE);
	}
	std::cout<<"-> \""<<playername<<"\""<<std::endl;

	/*
		Resolution selection
	*/

	u16 screenW;
	u16 screenH;
	bool fullscreen = false;
	
	if(g_screenW != "" && g_screenH != "")
	{
		screenW = atoi(g_screenW.c_str());
		screenH = atoi(g_screenH.c_str());
	}
	else
	{
		u16 resolutions[][3] = {
			//W, H, fullscreen
			{640,480, 0},
			{800,600, 0},
			{1024,768, 0},
			{1280,1024, 0},
			/*{640,480, 1},
			{800,600, 1},
			{1024,768, 1},
			{1280,1024, 1},*/
		};

		u16 res_count = sizeof(resolutions)/sizeof(resolutions[0]);
		
		for(u16 i=0; i<res_count; i++)
		{
			std::cout<<(i+1)<<": "<<resolutions[i][0]<<"x"
					<<resolutions[i][1];
			if(resolutions[i][2])
				std::cout<<" fullscreen"<<std::endl;
			else
				std::cout<<" windowed"<<std::endl;
		}
		std::cout<<"Select a window resolution number [empty = 2]: ";
		std::cin.getline(templine, 100);

		u16 r0;
		if(templine[0] == 0)
			r0 = 2;
		else
			r0 = atoi(templine);

		if(r0 > res_count || r0 == 0)
			r0 = 2;
		
		{
			u16 i = r0-1;
			std::cout<<"-> ";
			std::cout<<(i+1)<<": "<<resolutions[i][0]<<"x"
					<<resolutions[i][1];
			if(resolutions[i][2])
				std::cout<<" fullscreen"<<std::endl;
			else
				std::cout<<" windowed"<<std::endl;
		}

		screenW = resolutions[r0-1][0];
		screenH = resolutions[r0-1][1];
		fullscreen = resolutions[r0-1][2];
	}

	//

	video::E_DRIVER_TYPE driverType;

#ifdef _WIN32
	//driverType = video::EDT_DIRECT3D9; // Doesn't seem to work
	driverType = video::EDT_OPENGL;
#else
	driverType = video::EDT_OPENGL;
#endif

	// create device and exit if creation failed

	IrrlichtDevice *device;
	/*device = createDevice(driverType,
			core::dimension2d<u32>(screenW, screenH),
			16, fullscreen, false, false, &receiver);*/
	// With vsync
	device = createDevice(driverType,
			core::dimension2d<u32>(screenW, screenH),
			16, fullscreen, false, true, &receiver);

	if (device == 0)
		return 1; // could not create selected driver.
	
	device->setResizable(true);
	
	/*
		Run some speed tests
	*/
	//SpeedTests(device);

	/*
		Continue initialization
	*/

	video::IVideoDriver* driver = device->getVideoDriver();
	// These make the textures not to show at all
	//driver->setTextureCreationFlag(video::ETCF_ALWAYS_16_BIT);
	//driver->setTextureCreationFlag(video::ETCF_OPTIMIZED_FOR_SPEED );

	scene::ISceneManager* smgr = device->getSceneManager();

	gui::IGUIEnvironment* guienv = device->getGUIEnvironment();
	gui::IGUISkin* skin = guienv->getSkin();
	gui::IGUIFont* font = guienv->getFont("../data/fontlucida.png");
	if(font)
		skin->setFont(font);
	//skin->setColor(gui::EGDC_BUTTON_TEXT, video::SColor(255,0,0,0));
	skin->setColor(gui::EGDC_BUTTON_TEXT, video::SColor(255,255,255,255));
	//skin->setColor(gui::EGDC_3D_HIGH_LIGHT, video::SColor(0,0,0,0));
	//skin->setColor(gui::EGDC_3D_SHADOW, video::SColor(0,0,0,0));
	skin->setColor(gui::EGDC_3D_HIGH_LIGHT, video::SColor(255,0,0,0));
	skin->setColor(gui::EGDC_3D_SHADOW, video::SColor(255,0,0,0));
	
	const wchar_t *text = L"Loading and connecting...";
	core::vector2d<s32> center(screenW/2, screenH/2);
	core::dimension2d<u32> textd = font->getDimension(text);
	std::cout<<DTIME<<"Text w="<<textd.Width<<" h="<<textd.Height<<std::endl;
	// Have to add a bit to disable the text from word wrapping
	//core::vector2d<s32> textsize(textd.Width+4, textd.Height);
	core::vector2d<s32> textsize(300, textd.Height);
	core::rect<s32> textrect(center - textsize/2, center + textsize/2);

	gui::IGUIStaticText *gui_loadingtext = guienv->addStaticText(
			text, textrect, false, false);
	gui_loadingtext->setTextAlignment(gui::EGUIA_CENTER, gui::EGUIA_UPPERLEFT);

	driver->beginScene(true, true, video::SColor(255,0,0,0));
	guienv->drawAll();
	driver->endScene();

	/*
		Initialize material array
	*/

	//video::SMaterial g_materials[MATERIALS_COUNT];
	for(u16 i=0; i<MATERIALS_COUNT; i++)
	{
		g_materials[i].Lighting = false;
		g_materials[i].BackfaceCulling = false;

		const char *filename = g_material_filenames[i];
		if(filename != NULL){
			video::ITexture *t = driver->getTexture(filename);
			if(t == NULL){
				std::cout<<DTIME<<"Texture could not be loaded: \""
						<<filename<<"\""<<std::endl;
				return 1;
			}
			g_materials[i].setTexture(0, driver->getTexture(filename));
		}
		//g_materials[i].setFlag(video::EMF_TEXTURE_WRAP, video::ETC_REPEAT);
		g_materials[i].setFlag(video::EMF_BILINEAR_FILTER, false);
		//g_materials[i].setFlag(video::EMF_ANISOTROPIC_FILTER, false);
	}

	// Make a scope here for the client so that it gets removed
	// before the irrlicht device
	{

	std::cout<<DTIME<<"Creating server and client"<<std::endl;
	
	/*
		Create server
	*/
	SharedPtr<Server> server;
	if(hosting){
		server = new Server("../map");
		server->start(port);
	}
	
	/*
		Create client
	*/

	dout_client_ptr = &std::cout;

	// TODO: Get rid of the g_materials parameter or it's globalness
	Client client(device, g_materials);
	
	Player *player = client.getLocalPlayer();
	player->updateName(playername);
	
	Address connect_address(0,0,0,0, port);
	try{
		connect_address.Resolve(connect_name);
	}
	catch(ResolveError &e)
	{
		std::cout<<DTIME<<"Couldn't resolve address"<<std::endl;
		return 0;
	}
	
	std::cout<<DTIME<<"Connecting to server..."<<std::endl;
	client.connect(connect_address);
	
	try{
		while(client.connectedAndInitialized() == false)
		{
			client.step(0.1);
			if(server != NULL){
				server->step(0.1);
			}
			sleep_ms(100);
		}
	}
	catch(con::PeerNotFoundException &e)
	{
		std::cout<<DTIME<<"Timed out."<<std::endl;
#ifdef _WIN32
		std::cout<<DTIME<<"Press enter to continue... "<<std::endl;
		std::cin.get();
#endif
		return 0;
	}
	
	/*
		Create the camera node
	*/

	scene::ICameraSceneNode* camera = smgr->addCameraSceneNode(
		0, // Camera parent
		v3f(BS*100, BS*2, BS*100), // Look from
		v3f(BS*100+1, BS*2, BS*100), // Look to
		-1 // Camera ID
   	);

	if(camera == NULL)
		return 1;
	
	camera->setFOV(FOV_ANGLE);
	// Just so big a value that everything rendered is visible
	camera->setFarValue(BS*1000);
	
	f32 camera_yaw = 0; // "right/left"
	f32 camera_pitch = 0; // "up/down"
	
	// Random constants
#define WALK_ACCELERATION (4.0 * BS)
#define WALKSPEED_MAX (4.0 * BS)
//#define WALKSPEED_MAX (20.0 * BS)
	f32 walk_acceleration = WALK_ACCELERATION;
	f32 walkspeed_max = WALKSPEED_MAX;
	
	/*
		The mouse cursor needs not be visible, so we hide it via the
		irr::IrrlichtDevice::ICursorControl.
	*/
	device->getCursorControl()->setVisible(false);
	
	gui_loadingtext->remove();
	
	gui::IGUIStaticText *guitext = guienv->addStaticText(
			L"Minetest-c55",
			core::rect<s32>(5, 5, 5+600, 5+textsize.Y),
			false, false);
	gui::IGUIStaticText *guitext2 = guienv->addStaticText(
			L"",
			core::rect<s32>(5, 5+(textsize.Y+5)*1, 5+600, (5+textsize.Y)*2),
			false, false);
	/*
		Main loop
	*/

	bool first_loop_after_window_activation = true;
	
	s32 lastFPS = -1;
	
	// Time is in milliseconds
	u32 lasttime = device->getTimer()->getTime();
	
	while(device->run())
	{
		/*
			FPS limiter
		*/
		/*{
			u32 time = device->getTimer()->getTime();
			u32 frametime_min = 1000./50;
			if(time > framerate_min)
				device->sleep(framerate_min - time);
		}*/
		/*
			Time difference calculation
		*/
		u32 time = device->getTimer()->getTime();
		f32 dtime; // in seconds
		if(time > lasttime)
			dtime = (time - lasttime) / 1000.0;
		else
			dtime = 0;
		lasttime = time;

		static f32 dtime_avg1 = 0.0;
		dtime_avg1 = dtime_avg1 * 0.99 + dtime * 0.01;
		f32 jitter1 = dtime - dtime_avg1;
		
		static f32 jitter1_max_sample = 0.0;
		static f32 jitter1_min_sample = 0.0;
		{
			static f32 jitter1_max = 0.0;
			static f32 jitter1_min = 0.0;
			static f32 counter = 0.0;
			if(jitter1 > jitter1_max)
				jitter1_max = jitter1;
			if(jitter1 < jitter1_min)
				jitter1_min = jitter1;
			counter += dtime;
			if(counter > 0.0){
				counter -= 3.0;
				jitter1_max_sample = jitter1_max;
				jitter1_min_sample = jitter1_min;
				jitter1_max = 0.0;
				jitter1_min = 0.0;
			}
		}

		// Let's test with some jitter!
		/*{
			static s32 counter=0;
			if(counter > 0)
				counter--;
			else{
				counter = rand()%40;
				sleep_ms(rand()%500);
			}
		}*/

		/*
			Debug info for client
		*/
		{
			static float counter = 0.0;
			counter -= dtime;
			if(counter < 0)
			{
				counter = 30.0;
				client.printDebugInfo(std::cout);
			}
		}

		/*
			Viewing range
		*/
		
		updateViewingRange(dtime, &client);
		

		// Hilight boxes collected during the loop and displayed
		core::list< core::aabbox3d<f32> > hilightboxes;
		
		/*
			Special keys
		*/
		//if(receiver.IsKeyDown(irr::KEY_ESCAPE))
		if(g_esc_pressed)
		{
			break;
		}

		/*
			Player speed control
		*/

		v3f move_direction = v3f(0,0,1);
		move_direction.rotateXZBy(camera_yaw);
		
		v3f speed = v3f(0,0,0);
		if(receiver.IsKeyDown(irr::KEY_KEY_W))
		{
			speed += move_direction;
		}
		if(receiver.IsKeyDown(irr::KEY_KEY_S))
		{
			speed -= move_direction;
		}
		if(receiver.IsKeyDown(irr::KEY_KEY_A))
		{
			speed += move_direction.crossProduct(v3f(0,1,0));
		}
		if(receiver.IsKeyDown(irr::KEY_KEY_D))
		{
			speed += move_direction.crossProduct(v3f(0,-1,0));
		}
		if(receiver.IsKeyDown(irr::KEY_SPACE))
		{
			if(player->touching_ground){
				//player_speed.Y = 30*BS;
				//player.speed.Y = 5*BS;
				player->speed.Y = 6.5*BS;
			}
		}

		// The speed of the player (Y is ignored)
		speed = speed.normalize() * walkspeed_max;
		
		f32 inc = walk_acceleration * BS * dtime;
		
		if(player->speed.X < speed.X - inc)
			player->speed.X += inc;
		else if(player->speed.X > speed.X + inc)
			player->speed.X -= inc;
		else if(player->speed.X < speed.X)
			player->speed.X = speed.X;
		else if(player->speed.X > speed.X)
			player->speed.X = speed.X;

		if(player->speed.Z < speed.Z - inc)
			player->speed.Z += inc;
		else if(player->speed.Z > speed.Z + inc)
			player->speed.Z -= inc;
		else if(player->speed.Z < speed.Z)
			player->speed.Z = speed.Z;
		else if(player->speed.Z > speed.Z)
			player->speed.Z = speed.Z;
		
		/*
			Process environment
		*/
		
		{
			//TimeTaker("client.step(dtime)", device);
			client.step(dtime);
		}

		if(server != NULL){
			//TimeTaker("server->step(dtime)", device);
			server->step(dtime);
		}
		
		/*
			Mouse and camera control
		*/
		
		if(device->isWindowActive())
		{
			if(first_loop_after_window_activation){
				first_loop_after_window_activation = false;
			}
			else{
				s32 dx = device->getCursorControl()->getPosition().X - 320;
				s32 dy = device->getCursorControl()->getPosition().Y - 240;
				camera_yaw -= dx*0.2;
				camera_pitch += dy*0.2;
				if(camera_pitch < -89.9) camera_pitch = -89.9;
				if(camera_pitch > 89.9) camera_pitch = 89.9;
			}
			device->getCursorControl()->setPosition(320, 240);
		}
		else{
			first_loop_after_window_activation = true;
		}
		
		v3f camera_direction = v3f(0,0,1);
		camera_direction.rotateYZBy(camera_pitch);
		camera_direction.rotateXZBy(camera_yaw);

		v3f camera_position =
				player->getPosition() + v3f(0, BS+BS/2, 0);

		camera->setPosition(camera_position);
		camera->setTarget(camera_position + camera_direction);

		if(FIELD_OF_VIEW_TEST){
			//client.m_env.getMap().updateCamera(v3f(0,0,0), v3f(0,0,1));
			client.updateCamera(v3f(0,0,0), v3f(0,0,1));
		}
		else{
			//client.m_env.getMap().updateCamera(camera_position, camera_direction);
			client.updateCamera(camera_position, camera_direction);
		}
		
		/*
			Calculate what block is the crosshair pointing to
		*/
		
		//u32 t1 = device->getTimer()->getTime();
		
		//f32 d = 4; // max. distance
		f32 d = 4; // max. distance
		core::line3d<f32> shootline(camera_position,
				camera_position + camera_direction * BS * (d+1));
		
		bool nodefound = false;
		v3s16 nodepos;
		v3s16 neighbourpos;
		core::aabbox3d<f32> nodefacebox;
		f32 mindistance = BS * 1001;
		
		v3s16 pos_i = Map::floatToInt(player->getPosition());

		s16 a = d;
		s16 ystart = pos_i.Y + 0 - (camera_direction.Y<0 ? a : 1);
		s16 zstart = pos_i.Z - (camera_direction.Z<0 ? a : 1);
		s16 xstart = pos_i.X - (camera_direction.X<0 ? a : 1);
		s16 yend = pos_i.Y + 1 + (camera_direction.Y>0 ? a : 1);
		s16 zend = pos_i.Z + (camera_direction.Z>0 ? a : 1);
		s16 xend = pos_i.X + (camera_direction.X>0 ? a : 1);
		
		for(s16 y = ystart; y <= yend; y++){
		for(s16 z = zstart; z <= zend; z++){
		for(s16 x = xstart; x <= xend; x++)
		{
			try{
				//if(client.m_env.getMap().getNode(x,y,z).d == MATERIAL_AIR){
				if(client.getNode(v3s16(x,y,z)).d == MATERIAL_AIR){
					continue;
				}
			}catch(InvalidPositionException &e){
				continue;
			}

			v3s16 np(x,y,z);
			v3f npf = Map::intToFloat(np);
			
			f32 d = 0.01;
			
			v3s16 directions[6] = {
				v3s16(0,0,1), // back
				v3s16(0,1,0), // top
				v3s16(1,0,0), // right
				v3s16(0,0,-1),
				v3s16(0,-1,0),
				v3s16(-1,0,0),
			};

			for(u16 i=0; i<6; i++){
			//{u16 i=3;
				v3f dir_f = v3f(directions[i].X,
						directions[i].Y, directions[i].Z);
				v3f centerpoint = npf + dir_f * BS/2;
				f32 distance =
						(centerpoint - camera_position).getLength();
				
				if(distance < mindistance){
					//std::cout<<DTIME<<"for centerpoint=("<<centerpoint.X<<","<<centerpoint.Y<<","<<centerpoint.Z<<"): distance < mindistance"<<std::endl;
					//std::cout<<DTIME<<"npf=("<<npf.X<<","<<npf.Y<<","<<npf.Z<<")"<<std::endl;
					core::CMatrix4<f32> m;
					m.buildRotateFromTo(v3f(0,0,1), dir_f);

					// This is the back face
					v3f corners[2] = {
						v3f(BS/2, BS/2, BS/2),
						v3f(-BS/2, -BS/2, BS/2+d)
					};
					
					for(u16 j=0; j<2; j++){
						m.rotateVect(corners[j]);
						corners[j] += npf;
						//std::cout<<DTIME<<"box corners["<<j<<"]: ("<<corners[j].X<<","<<corners[j].Y<<","<<corners[j].Z<<")"<<std::endl;
					}

					//core::aabbox3d<f32> facebox(corners[0],corners[1]);
					core::aabbox3d<f32> facebox(corners[0]);
					facebox.addInternalPoint(corners[1]);

					if(facebox.intersectsWithLine(shootline)){
						nodefound = true;
						nodepos = np;
						neighbourpos = np + directions[i];
						mindistance = distance;
						nodefacebox = facebox;
					}
				}
			}
		}}}

		if(nodefound){
			//std::cout<<DTIME<<"nodefound == true"<<std::endl;
			//std::cout<<DTIME<<"nodepos=("<<nodepos.X<<","<<nodepos.Y<<","<<nodepos.Z<<")"<<std::endl;
			//std::cout<<DTIME<<"neighbourpos=("<<neighbourpos.X<<","<<neighbourpos.Y<<","<<neighbourpos.Z<<")"<<std::endl;

			static v3s16 nodepos_old(-1,-1,-1);
			if(nodepos != nodepos_old){
				std::cout<<DTIME<<"Pointing at ("<<nodepos.X<<","
						<<nodepos.Y<<","<<nodepos.Z<<")"<<std::endl;
				nodepos_old = nodepos;

				/*wchar_t positiontext[20];
				swprintf(positiontext, 20, L"(%i,%i,%i)",
						nodepos.X, nodepos.Y, nodepos.Z);
				positiontextgui->setText(positiontext);*/
			}

			hilightboxes.push_back(nodefacebox);
			
			if(receiver.leftclicked){
				std::cout<<DTIME<<"Removing node"<<std::endl;
				client.removeNode(nodepos);
			}
			if(receiver.rightclicked){
				std::cout<<DTIME<<"Placing node"<<std::endl;

				MapNode n;
				n.d = g_selected_material;
				client.addNode(neighbourpos, n);
			}
		}
		else{
			//std::cout<<DTIME<<"nodefound == false"<<std::endl;
			//positiontextgui->setText(L"");
		}

		receiver.leftclicked = false;
		receiver.rightclicked = false;

		/*
			Update gui stuff
		*/
		
		static u8 old_selected_material = MATERIAL_AIR;
		static bool old_vr_all = 0;
		if(g_selected_material != old_selected_material
				|| g_viewing_range_all != old_vr_all)
		{
			old_selected_material = g_selected_material;
			old_vr_all = g_viewing_range_all;
			wchar_t temptext[100];
			swprintf(temptext, 100, L"Minetest-c55 ("
					L"F: material=%i"
					L", R: range_all=%i"
					L")",
					g_selected_material,
					g_viewing_range_all);
			guitext->setText(temptext);
		}
		
		{
			wchar_t temptext[100];
			swprintf(temptext, 100,
					L"(% .4f < frametime_jitter < % .4f)",
					jitter1_min_sample,
					jitter1_max_sample);
			guitext2->setText(temptext);
		}

		/*
			Calculate stuff for drawing
		*/

		v2u32 screensize = driver->getScreenSize();
		core::vector2d<s32> displaycenter(screensize.X/2,screensize.Y/2);

		camera->setAspectRatio((f32)screensize.X / (f32)screensize.Y);
		
		/*
			Drawing begins
		*/

		/*
			Background color is choosen based on whether the player is
			much beyond the initial ground level
		*/
		/*video::SColor bgcolor;
		v3s16 p0 = Map::floatToInt(player->getPosition());
		// Does this make short random delays?
		// NOTE: no need for this, sky doesn't show underground with
		// enough range
		bool is_underground = client.isNodeUnderground(p0);
		//bool is_underground = false;
		if(is_underground == false)
			bgcolor = video::SColor(255,90,140,200);
		else
			bgcolor = video::SColor(255,0,0,0);*/
			
		video::SColor bgcolor = video::SColor(255,90,140,200);

		driver->beginScene(true, true, bgcolor);

		//std::cout<<DTIME<<"smgr->drawAll()"<<std::endl;

		smgr->drawAll();

		driver->draw2DLine(displaycenter - core::vector2d<s32>(10,0),
				displaycenter + core::vector2d<s32>(10,0),
				video::SColor(255,255,255,255));
		driver->draw2DLine(displaycenter - core::vector2d<s32>(0,10),
				displaycenter + core::vector2d<s32>(0,10),
				video::SColor(255,255,255,255));
		
		video::SMaterial m;
		m.Thickness = 10;
		m.Lighting = false;
		driver->setMaterial(m);

		for(core::list< core::aabbox3d<f32> >::Iterator i=hilightboxes.begin();
				i != hilightboxes.end(); i++){
			driver->draw3DBox(*i, video::SColor(255,0,0,0));
		}

		guienv->drawAll();

		driver->endScene();

		/*
			Drawing ends
		*/

		u16 fps = driver->getFPS();

		if (lastFPS != fps)
		{
			core::stringw str = L"Minetest [";
			str += driver->getName();
			str += "] FPS:";
			str += fps;

			device->setWindowCaption(str.c_str());
			lastFPS = fps;
		}


		/*}
		else
			device->yield();*/
	}

	} // client is deleted at this point

	/*
	In the end, delete the Irrlicht device.
	*/
	device->drop();

	} //try
	catch(con::PeerNotFoundException &e)
	{
		std::cout<<DTIME<<"Connection timed out."<<std::endl;
#ifdef _WIN32
		std::cout<<DTIME<<"Press enter to continue... "<<std::endl;
		std::cin.get();
#endif
	}
	/*catch(std::exception &e)
	{
		std::cout<<DTIME<<"An unhandled exception occurred: "
				<<e.what()<<std::endl;
		std::cout<<DTIME<<"Quitting. Please see "DEBUGFILE
				" for some information."<<std::endl;
#ifdef _WIN32
		std::cout<<DTIME<<"Press enter to continue... "<<std::endl;
		std::cin.get();
#endif
	}*/
	
	return 0;
}

//END
