#ifndef SERIALIZATION_HEADER
#define SERIALIZATION_HEADER

#include "common_irrlicht.h"
#include "exceptions.h"
#include <iostream>
#include "utility.h"

/*
	Serialization format versions:
	0: original networked test with 1-byte nodes
	1: update with 2-byte nodes
	2: lighting is transmitted in param
	3: optional fetching of far blocks
	4: block compression
	5: sector objects NOTE: block compression was left accidentally out
	6: failed attempt at switching block compression on again
	7: block compression switched on again
*/
// This represents an uninitialized or invalid format
#define SER_FMT_VER_INVALID 255
// Highest supported serialization version
#define SER_FMT_VER_HIGHEST 7
// Lowest supported serialization version
#define SER_FMT_VER_LOWEST 0

#define ser_ver_supported(v) (v >= SER_FMT_VER_LOWEST && v <= SER_FMT_VER_HIGHEST)

void compress(SharedBuffer<u8> data, std::ostream &os, u8 version);
void decompress(std::istream &is, std::ostream &os, u8 version);

// There is no benefits in using this kind of thing and it
// makes using serializedLength as it can't be static.
/*class Serializable
{
public:
	virtual u32 serializedLength(u8 version)
	{
		throw NotImplementedException("serializedLength() stub called");
	}
	virtual void serialize(u8 *dest, u8 version)
	{
		throw NotImplementedException("serialize() stub called");
	}
	virtual void deSerialize(u8 *source, u8 version)
	{
		throw NotImplementedException("deSerialize() stub called");
	}
};*/

#endif

