/*
(c) 2010 Perttu Ahola <celeron55@gmail.com>
*/

#ifndef INVENTORY_HEADER
#define INVENTORY_HEADER

#include <iostream>
#include <string>
#include "common_irrlicht.h"
#include "debug.h"

class InventoryItem
{
public:
	InventoryItem();
	virtual ~InventoryItem();
	
	static InventoryItem* deSerialize(std::istream &is);
	
	virtual const char* getName() const = 0;
	// Shall write the name and the parameters
	virtual void serialize(std::ostream &os) = 0;
	// Shall make an exact clone of the item
	virtual InventoryItem* clone() = 0;

private:
};

#define MATERIAL_ITEM_MAX_COUNT 99

class MaterialItem : public InventoryItem
{
public:
	MaterialItem(u8 material, u16 count)
	{
		m_material = material;
		m_count = count;
	}
	virtual const char* getName() const
	{
		return "MaterialItem";
	}
	virtual void serialize(std::ostream &os)
	{
		//os.imbue(std::locale("C"));
		os<<getName();
		os<<" ";
		os<<(unsigned int)m_material;
		os<<" ";
		os<<m_count;
	}
	virtual InventoryItem* clone()
	{
		return new MaterialItem(m_material, m_count);
	}

	u8 getMaterial()
	{
		return m_material;
	}
	u16 getCount()
	{
		return m_count;
	}
	u16 freeSpace()
	{
		return MATERIAL_ITEM_MAX_COUNT - m_count;
	}
	void add(u16 count)
	{
		assert(m_count + count <= MATERIAL_ITEM_MAX_COUNT);
		m_count += count;
	}
	void remove(u16 count)
	{
		assert(m_count >= count);
		m_count -= count;
	}
private:
	u8 m_material;
	u16 m_count;
};

class Inventory
{
public:
	Inventory(u32 size);
	~Inventory();
	void clearItems();
	void serialize(std::ostream &os);
	void deSerialize(std::istream &is);

	Inventory & operator = (Inventory &other);

	u32 getSize();
	InventoryItem * getItem(u32 i);
	// Returns old item (or NULL). Parameter can be NULL.
	InventoryItem * changeItem(u32 i, InventoryItem *newitem);
	void deleteItem(u32 i);
	// Adds an item to a suitable place. Returns false if failed.
	bool addItem(InventoryItem *newitem);

	void print(std::ostream &o);
	
private:
	core::array<InventoryItem*> m_items;
	u32 m_size;
};

#endif

